#!/usr/bin/env python3

import os
from optparse import OptionParser
from sys import exit, version_info

from string_extractor.parse import parse_json_file
from string_extractor.pot_export import write_to_pot, sanitize


parser = OptionParser()
parser.add_option("-i", "--include_dir", dest="include_dir",
                  action="append", type="str",
                  help="include directories")
parser.add_option("-n", "--name", dest="name", help="POT package name")
parser.add_option("-r", "--reference", dest="reference",
                  help="reference POT for plural collision avoidance, "
                  "also strings from JSON are appended to this file")
parser.add_option("-v", "--verbose", dest="verbose", help="be verbose")
parser.add_option("-X", "--exclude", dest="exclude",
                  action="append", type="str",
                  help="exclude a single file")
parser.add_option("-x", "--exclude_dir", dest="exclude_dir",
                  action="append", type="str",
                  help="exclude directories")
parser.add_option("-D", "--obsolete_paths", dest="obsolete_paths",
                  action="append", type="str",
                  help="obsolete directories or files")

(options, args) = parser.parse_args()

if not (version_info.major >= 3 and version_info.minor >= 7):
    print("Requires Python 3.7 or higher.")
    exit(1)

if not options.reference:
    print("Have to specify reference file path.")
    exit(1)

if not options.include_dir:
    print("Have to specify at least one search path.")
    exit(1)

include_dir = [os.path.normpath(i) for i in options.include_dir]

exclude = [os.path.normpath(i) for i in options.exclude] \
    if options.exclude else []

exclude_dir = [os.path.normpath(i) for i in options.exclude_dir] \
    if options.exclude_dir else []

obsolete_paths = [os.path.normpath(i) for i in options.obsolete_paths] \
    if options.obsolete_paths else []


def extract_all_from_dir(json_dir):
    """Extract strings from every json file in the specified directory,
    recursing into any subdirectories."""
    allfiles = sorted(os.listdir(json_dir))
    dirs = []
    skiplist = [os.path.normpath(".gitkeep")]
    for f in allfiles:
        full_name = os.path.join(json_dir, f)
        if full_name in [i for i in include_dir if i != json_dir]:
            # Skip other included directories;
            # They will be extracted later and appended to
            # the end of the shared list of strings;
            continue
        if os.path.isdir(full_name):
            dirs.append(f)
        elif f in skiplist or full_name in exclude:
            continue
        elif any(full_name.startswith(dir) for dir in exclude_dir):
            continue
        elif f.endswith(".json"):
            parse_json_file(full_name)
    for d in dirs:
        extract_all_from_dir(os.path.join(json_dir, d))


def main():
    for i in include_dir:
        extract_all_from_dir(i)

    sanitize(options.reference, options.name)

    with open(options.reference, mode="a", encoding="utf-8") as fp:
        write_to_pot(fp, obsolete_paths=obsolete_paths)


main()



from .enchant import parse_enchant
from ..helper import get_singular_name
from ..write_text import write_text


def parse_effect_type(json, origin):
    # Using a list to reserve order and ensure stability across runs
    effect_name = []

    if "name" in json:
        for name in json["name"]:
            write_text(name, origin,
                       comment="Name of effect type id \"{}\""
                       .format(json["id"]))
            singular = get_singular_name(name)
            if len(singular) > 0 and singular not in effect_name:
                effect_name.append(singular)

    if len(effect_name) > 0:
        effect_name = ", ".join(effect_name)
    else:
        effect_name = json["id"]

    if "desc" in json:
        # See effect.cpp
        use_desc_ints = json.get("max_intensity", 1) <= len(json["desc"])
        for idx, desc in enumerate(json["desc"]):
            if use_desc_ints:
                names = json.get("name", [])
                if len(names) > 0:
                    singular = get_singular_name(
                        names[min(len(names) - 1, idx)])
                    if len(singular) == 0:
                        singular = json["id"]
                else:
                    singular = json["id"]
            else:
                singular = effect_name
            write_text(desc, origin,
                       comment="Description of effect type \"{}\""
                       .format(singular))

    if "speed_name" in json:
        write_text(json["speed_name"], origin,
                   comment="Speed name of effect type \"{}\""
                   .format(effect_name))

    if "apply_message" in json:
        if type(json["apply_message"]) is list:
            for msg in json["apply_message"]:
                write_text(msg[0], origin,
                           comment="Apply message of effect type \"{}\""
                           .format(effect_name))
        elif type(json["apply_message"]) is str:
            write_text(json["apply_message"], origin,
                       comment="Apply message of effect type \"{}\""
                       .format(effect_name))

    if "remove_message" in json:
        write_text(json["remove_message"], origin,
                   comment="Remove message of effect type \"{}\""
                   .format(effect_name))

    if "death_msg" in json:
        write_text(json["death_msg"], origin,
                   comment="Death message of effect type \"{}\""
                   .format(effect_name))

    if "miss_messages" in json:
        for msg in json["miss_messages"]:
            write_text(msg[0], origin,
                       comment="Miss message of effect type \"{}\""
                       .format(effect_name))

    if "decay_messages" in json:
        for msg in json["decay_messages"]:
            write_text(msg[0], origin,
                       comment="Decay message of effect type \"{}\""
                       .format(effect_name))

    if "apply_memorial_log" in json:
        write_text(json["apply_memorial_log"], origin,
                   context="memorial_male",
                   comment="Male memorial apply log of effect type \"{}\""
                   .format(effect_name))
        write_text(json["apply_memorial_log"], origin,
                   context="memorial_female",
                   comment="Female memorial apply log of effect type \"{}\""
                   .format(effect_name))

    if "remove_memorial_log" in json:
        write_text(json["remove_memorial_log"], origin,
                   context="memorial_male",
                   comment="Male memorial remove log of effect type \"{}\""
                   .format(effect_name))
        write_text(json["remove_memorial_log"], origin,
                   context="memorial_female",
                   comment="Female memorial remove log of effect type \"{}\""
                   .format(effect_name))

    if "blood_analysis_description" in json:
        write_text(json["blood_analysis_description"], origin,
                   comment="Blood analysis description of effect type \"{}\""
                   .format(effect_name))

    if "enchantments" in json:
        for enchantment in json["enchantments"]:
            parse_enchant(enchantment, origin)



from ..helper import get_singular_name
from ..write_text import write_text


def parse_mod_info(json, origin):
    name = get_singular_name(json["name"])
    write_text(json["name"], origin, comment="MOD name", c_format=False)
    write_text(json["description"], origin, c_format=False,
               comment="Description of MOD \"{}\"".format(name))



# -*- coding: utf-8 -*-


import biplist
import os.path

# defines is provided by dmgbuild, but pretend to define it here for the
# beneit of linting tools.
defines = defines  # noqa: F821

# .. Useful stuff .............................................................

application = defines.get('app', 'Cataclysm.app')
appname = os.path.basename(application)


def icon_from_app(app_path):
    plist_path = os.path.join(app_path, 'Contents', 'Info.plist')
    plist = biplist.readPlist(plist_path)
    icon_name = plist['CFBundleIconFile']
    icon_root, icon_ext = os.path.splitext(icon_name)
    if not icon_ext:
        icon_ext = '.icns'
    icon_name = icon_root + icon_ext
    return os.path.join(app_path, 'Contents', 'Resources', icon_name)


# .. Basics ...................................................................

# Uncomment to override the output filename
# filename = 'test.dmg'

# Uncomment to override the output volume name
# volume_name = 'Test'

# Volume format (see hdiutil create -help)
format = defines.get('format', 'UDBZ')

# Volume size (must be large enough for your files)
size = defines.get('size', '500M')

# Files to include
files = [application]

# Symlinks to create
symlinks = {'Applications': '/Applications'}

# Volume icon
#
# You can either define icon, in which case that icon file will be copied to
# the image, *or* you can define badge_icon, in which case the icon file you
# specify will be used to badge the system's Removable Disk icon
#
#icon = '/path/to/icon.icns'
badge_icon = icon_from_app(application)

# Where to put the icons
icon_locations = {
    appname: (240, 400),
    'Applications': (430, 400)
}

# .. Window configuration .....................................................

# Background
#
# This is a STRING containing any of the following:
#
#    #3344ff          - web-style RGB color
#    #34f             - web-style RGB color, short form (#34f == #3344ff)
#    rgb(1,0,0)       - RGB color, each value is between 0 and 1
#    hsl(120,1,.5)    - HSL (hue saturation lightness) color
#    hwb(300,0,0)     - HWB (hue whiteness blackness) color
#    cmyk(0,1,0,0)    - CMYK color
#    goldenrod        - X11/SVG named color
#    builtin-arrow    - A simple built-in background with a blue arrow
#    /foo/bar/baz.png - The path to an image file
#
# The hue component in hsl() and hwb() may include a unit; it defaults to
# degrees ('deg'), but also supports radians ('rad') and gradians ('grad'
# or 'gon').
#
# Other color components may be expressed either in the range 0 to 1, or
# as percentages (e.g. 60% is equivalent to 0.6).
background = 'build-data/osx/dmgback.png'

show_status_bar = False
show_tab_view = False
show_toolbar = False
show_pathbar = False
show_sidebar = False
sidebar_width = 180

# Window position in ((x, y), (w, h)) format
window_rect = ((200, 160), (640, 503))

# Select the default view; must be one of
#
#    'icon-view'
#    'list-view'
#    'column-view'
#    'coverflow'
#
default_view = 'icon-view'

# General view configuration
show_icon_preview = False

# Set these to True to force inclusion of icon/list view settings (otherwise
# we only include settings for the default view)
include_icon_view_settings = 'auto'
include_list_view_settings = 'auto'

# .. Icon view configuration ..................................................

arrange_by = None
grid_offset = (0, 0)
grid_spacing = 100
scroll_position = (0, 0)
label_pos = 'bottom'  # or 'right'
text_size = 16
icon_size = 64

# .. List view configuration ..................................................

# Column names are as follows:
#
#   name
#   date-modified
#   date-created
#   date-added
#   date-last-opened
#   size
#   kind
#   label
#   version
#   comments
#
list_icon_size = 16
list_text_size = 12
list_scroll_position = (0, 0)
list_sort_by = 'name'
list_use_relative_dates = True
list_calculate_all_sizes = False,
list_columns = ('name', 'date-modified', 'size', 'kind', 'date-added')
list_column_widths = {
    'name': 300,
    'date-modified': 181,
    'date-created': 181,
    'date-added': 181,
    'date-last-opened': 181,
    'size': 97,
    'kind': 115,
    'label': 100,
    'version': 75,
    'comments': 300,
}
list_column_sort_directions = {
    'name': 'ascending',
    'date-modified': 'descending',
    'date-created': 'descending',
    'date-added': 'descending',
    'date-last-opened': 'descending',
    'size': 'descending',
    'kind': 'ascending',
    'label': 'ascending',
    'version': 'ascending',
    'comments': 'ascending',
}



from ..helper import get_singular_name
from ..write_text import write_text


def parse_help(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Help menu")

    for msg in json.get("messages", []):
        write_text(msg, origin,
                   comment="Help message in menu \"{}\"".format(name))



#!/usr/bin/env python3

# Usage 'extend_itemgroups.py dir file'
# Will convert all itemgroups within dir that have an id listed in 'file'
# to "copy-from" an itemgroup with the same id and use "extend" for all entries

import argparse
import json
import os


args = argparse.ArgumentParser()
args.add_argument("dir", action="store", help="specify json directory")
args.add_argument("converted", action="store",
                  help="specify file containing itemgroup ids."
                  "Only itemgroups with an id contained within"
                  "this file will be converted")
args_dict = vars(args.parse_args())

# The ids of itemgroups we convert
converted_ids = set()

# The keys we are moving to extend
transferred_keys = [
    "items",
    "entries",
    "groups"
]


def gen_new(path):
    change = False
    with open(path, "r", encoding="utf-8") as json_file:
        try:
            json_data = json.load(json_file)
        except json.JSONDecodeError:
            print(f"Json Decode Error at {path}")
            print("Ensure that the file is a JSON file consisting of"
                  " an array of objects!")
            return None

        for jo in json_data:
            if ("type" not in jo or jo["type"] != "item_group"):
                continue
            if jo["id"] not in converted_ids:
                continue

            change = True
            jo["copy-from"] = jo["id"]

            jo["extend"] = dict()
            for key in transferred_keys:
                if key not in jo:
                    continue
                jo["extend"][key] = jo[key]
                del jo[key]

        return json_data if change else None


all_filepaths = set()
for root, directories, filenames in os.walk(args_dict["dir"]):
    for filename in filenames:
        if not filename.endswith(".json"):
            continue

        path = os.path.join(root, filename)
        all_filepaths.add(path)

with open(args_dict["converted"]) as file:
    for line in file:
        # Remove the newline
        line = line[:-1]
        converted_ids.add(line)

for path in all_filepaths:
    new = gen_new(path)
    if new is not None:
        with open(path, "w", encoding="utf-8") as jf:
            json.dump(new, jf, ensure_ascii=False)
        os.system(f"./tools/format/json_formatter.cgi {path}")



#!/usr/bin/env python3
#
# Modified by the Cataclysm: Dark Days Ahead project.
#
#===- check_clang_tidy.py - ClangTidy Test Helper ------------*- python -*--===#
#
# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
#===------------------------------------------------------------------------===#

r"""
ClangTidy Test Helper
=====================

This script runs clang-tidy in fix mode and verify fixes, messages or both.

Usage:
  check_clang_tidy.py [-resource-dir=<resource-dir>] \
    [-assume-filename=<file-with-source-extension>] \
    [-check-suffix=<comma-separated-file-check-suffixes>] \
    [-check-suffixes=<comma-separated-file-check-suffixes>] \
    [-std=c++(98|11|14|17|20)[-or-later]] \
    [-clang-tidy-binary=<path/to/clang-tidy>] \
    <source-file> <check-name> <temp-file> \
    -- [optional clang-tidy arguments]

Example:
  // RUN: %check_clang_tidy %s llvm-include-order %t -- -- -isystem %S/Inputs

Notes:
  -std=c++(98|11|14|17|20)-or-later:
    This flag will cause multiple runs within the same check_clang_tidy
    execution. Make sure you don't have shared state across these runs.
  optional clang-tidy arguments:
    Passed to clang-tidy. Use another `--` to separate clang-tidy and clang
    arguments. Although optional clang-tidy arguments before the first `--`
    is supported, care should be taken as it may cause unexpected behavior for
    specific argument types and orders (https://github.com/CleverRaven/Cataclysm-DDA/pull/66315/files#r1247288210)
"""

import argparse
import os
import re
import subprocess
import sys


def write_file(file_name, text):
  with open(file_name, 'w', encoding='utf-8') as f:
    f.write(text)
    f.truncate()


def try_run(args, raise_error=True):
  try:
    process_output = \
      subprocess.check_output(args, stderr=subprocess.STDOUT).decode(errors='ignore')
  except subprocess.CalledProcessError as e:
    process_output = e.output.decode(errors='ignore')
    print('%s failed:\n%s' % (' '.join(args), process_output))
    if raise_error:
      raise
  return process_output


# This class represents the appearance of a message prefix in a file.
class MessagePrefix:
  def __init__(self, label):
    self.has_message = False
    self.prefixes = []
    self.label = label

  def check(self, file_check_suffix, input_text):
    self.prefix = self.label + file_check_suffix
    self.has_message = self.prefix in input_text
    if self.has_message:
      self.prefixes.append(self.prefix)
    return self.has_message


class CheckRunner:
  def __init__(self, args, extra_args):
    self.resource_dir = args.resource_dir
    self.assume_file_name = args.assume_filename
    self.input_file_name = args.input_file_name
    self.check_name = args.check_name
    self.temp_file_name = args.temp_file_name
    self.original_file_name = self.temp_file_name + ".orig"
    self.expect_clang_tidy_error = args.expect_clang_tidy_error
    self.std = args.std
    self.clang_tidy_binary = args.clang_tidy_binary
    self.check_suffix = args.check_suffix
    self.input_text = ''
    self.has_check_fixes = False
    self.has_check_messages = False
    self.has_check_notes = False
    self.fixes = MessagePrefix('CHECK-FIXES')
    self.messages = MessagePrefix('CHECK-MESSAGES')
    self.notes = MessagePrefix('CHECK-NOTES')

    file_name_with_extension = self.assume_file_name or self.input_file_name
    _, extension = os.path.splitext(file_name_with_extension)
    if extension not in ['.c', '.hpp', '.m', '.mm']:
      extension = '.cpp'
    self.temp_file_name = self.temp_file_name + extension

    self.clang_extra_args = []
    self.clang_tidy_extra_args = extra_args
    if '--' in extra_args:
      i = self.clang_tidy_extra_args.index('--')
      self.clang_extra_args = self.clang_tidy_extra_args[i + 1:]
      self.clang_tidy_extra_args = self.clang_tidy_extra_args[:i]

    # If the test does not specify a config style, force an empty one; otherwise
    # auto-detection logic can discover a ".clang-tidy" file that is not related to
    # the test.
    if not any(
        [arg.startswith('-config=') for arg in self.clang_tidy_extra_args]):
      self.clang_tidy_extra_args.append('-config={}')

    if extension in ['.m', '.mm']:
      self.clang_extra_args = ['-fobjc-abi-version=2', '-fobjc-arc', '-fblocks'] + \
          self.clang_extra_args

    if extension in ['.cpp', '.hpp', '.mm']:
      self.clang_extra_args.append('-std=' + self.std)

    # Tests should not rely on STL being available, and instead provide mock
    # implementations of relevant APIs.
    if not args.allow_stdinc:
      self.clang_extra_args.append('-nostdinc++')

    if self.resource_dir is not None:
      self.clang_extra_args.append('-resource-dir=%s' % self.resource_dir)

  def read_input(self):
    with open(self.input_file_name, 'r', encoding='utf-8') as input_file:
      self.input_text = input_file.read()

  def get_prefixes(self):
    for suffix in self.check_suffix:
      if suffix and not re.match('^[A-Z0-9\\-]+$', suffix):
        sys.exit('Only A..Z, 0..9 and "-" are allowed in check suffixes list,'
                 + ' but "%s" was given' % suffix)

      file_check_suffix = ('-' + suffix) if suffix else ''

      has_check_fix = self.fixes.check(file_check_suffix, self.input_text)
      self.has_check_fixes = self.has_check_fixes or has_check_fix

      has_check_message = self.messages.check(file_check_suffix, self.input_text)
      self.has_check_messages = self.has_check_messages or has_check_message

      has_check_note = self.notes.check(file_check_suffix, self.input_text)
      self.has_check_notes = self.has_check_notes or has_check_note

      if has_check_note and has_check_message:
        sys.exit('Please use either %s or %s but not both' %
          (self.notes.prefix, self.messages.prefix))

      if not has_check_fix and not has_check_message and not has_check_note:
        sys.exit('%s, %s or %s not found in the input' %
          (self.fixes.prefix, self.messages.prefix, self.notes.prefix))

    assert self.has_check_fixes or self.has_check_messages or self.has_check_notes

  def prepare_test_inputs(self):
    # Remove the contents of the CHECK lines to avoid CHECKs matching on
    # themselves.  We need to keep the comments to preserve line numbers while
    # avoiding empty lines which could potentially trigger formatting-related
    # checks.
    cleaned_test = re.sub('// *CHECK-[A-Z0-9\\-]*:[^\r\n]*', '//', self.input_text)
    write_file(self.temp_file_name, cleaned_test)
    write_file(self.original_file_name, cleaned_test)

  def run_clang_tidy(self):
    args = [self.clang_tidy_binary, self.temp_file_name, '-fix', '--checks=-*,' + self.check_name] + \
        self.clang_tidy_extra_args + ['--'] + self.clang_extra_args
    if self.expect_clang_tidy_error:
      args.insert(0, 'not')
    print('Running ' + repr(args) + '...')
    clang_tidy_output = try_run(args)
    print('------------------------ clang-tidy output -----------------------')
    print(clang_tidy_output.encode(sys.stdout.encoding, errors="replace").decode(sys.stdout.encoding))
    print('------------------------------------------------------------------')

    diff_output = try_run(['diff', '-u', self.original_file_name, self.temp_file_name], False)
    print('------------------------------ Fixes -----------------------------')
    print(diff_output)
    print('------------------------------------------------------------------')
    return clang_tidy_output

  def check_fixes(self):
    if self.has_check_fixes:
      try_run(['FileCheck', '-input-file=' + self.temp_file_name, self.input_file_name,
              '-check-prefixes=' + ','.join(self.fixes.prefixes),
              '-strict-whitespace'])

  def check_messages(self, clang_tidy_output):
    if self.has_check_messages:
      messages_file = self.temp_file_name + '.msg'
      write_file(messages_file, clang_tidy_output)
      try_run(['FileCheck', '-input-file=' + messages_file, self.input_file_name,
             '-check-prefixes=' + ','.join(self.messages.prefixes),
             '-implicit-check-not={{warning|error}}:'])

  def check_notes(self, clang_tidy_output):
    if self.has_check_notes:
      notes_file = self.temp_file_name + '.notes'
      filtered_output = [line for line in clang_tidy_output.splitlines()
                         if not ("note: FIX-IT applied" in line)]
      write_file(notes_file, '\n'.join(filtered_output))
      try_run(['FileCheck', '-input-file=' + notes_file, self.input_file_name,
             '-check-prefixes=' + ','.join(self.notes.prefixes),
             '-implicit-check-not={{note|warning|error}}:'])

  def run(self):
    self.read_input()
    self.get_prefixes()
    self.prepare_test_inputs()
    clang_tidy_output = self.run_clang_tidy()
    self.check_fixes()
    self.check_messages(clang_tidy_output)
    self.check_notes(clang_tidy_output)


def expand_std(std):
  if std == 'c++98-or-later':
    return ['c++98', 'c++11', 'c++14', 'c++17', 'c++20']
  if std == 'c++11-or-later':
    return ['c++11', 'c++14', 'c++17', 'c++20']
  if std == 'c++14-or-later':
    return ['c++14', 'c++17', 'c++20']
  if std == 'c++17-or-later':
    return ['c++17', 'c++20']
  if std == 'c++20-or-later':
    return ['c++20']
  return [std]


def csv(string):
  return string.split(',')


def parse_arguments():
  parser = argparse.ArgumentParser()
  parser.add_argument('-expect-clang-tidy-error', action='store_true')
  parser.add_argument('-resource-dir')
  parser.add_argument('-assume-filename')
  parser.add_argument('input_file_name')
  parser.add_argument('check_name')
  parser.add_argument('temp_file_name')
  parser.add_argument(
    '-check-suffix',
    '-check-suffixes',
    default=[''],
    type=csv,
    help='comma-separated list of FileCheck suffixes')
  parser.add_argument('-std', type=csv, default=['c++11-or-later'])
  parser.add_argument('-allow-stdinc', action='store_true')
  parser.add_argument('-clang-tidy-binary', default='clang-tidy')
  return parser.parse_known_args()


def main():
  args, extra_args = parse_arguments()

  abbreviated_stds = args.std
  for abbreviated_std in abbreviated_stds:
    for std in expand_std(abbreviated_std):
      args.std = std
      CheckRunner(args, extra_args).run()


if __name__ == '__main__':
  main()



#!/usr/bin/env python3
import os
import regex
from spellchecker import SpellChecker
import unicodedata


_latin_and_diacritic = (
    r"\p{Script=Latin}"
    r"|"
    r"[\u0300-\u036F\u1AB0-\u1AC0\u1DC0-\u1DFF\u20D0-\u20F0\uFE20-\uFE2D]"
)
Tokenizer = regex.compile(
    r"(?<="
        r"(?:"  # a word can be after...  # noqa: E131
            r"^"  # ...the start of text...  # noqa: E131
        r"|"  # ...or...
            r"""[\s\-'‘"“«\*&\(\[\{/>]"""  # ...a valid starting delimiter
        r")"
        r"[…]*"  # ignore possible punctuations before a word
    r")"
    # a single letter is always considered a valid word and therefore not
    # tokenized
    r"(?:" +
        _latin_and_diacritic +
    r")"  # a word always starts with an letter...
    r"(?:" +
        _latin_and_diacritic +
    r"|['‘’´])*"  # ...can contain letters and apostrophes...
    r"(?:" +
        _latin_and_diacritic +
    r")"  # ...and always ends with an letter
    # exclude suffix `'s` and match the apostrophe as a delimiter later
    r"(?<!'[sS])"
    # note: staring and ending apostrophe cannot be distinguished from
    # starting and ending single quote so the former is not considered.
    r"(?="
        r"[,\.\?!:;…]*"  # ignore any punctuations after a word
        r"(?:"  # a word can be before...
            r"$"  # ...the end of text...
        r"|"  # ...or...
            r"""[\s\-'’"”»\*\)\]\}/<]"""  # ...a valid ending delimiter
        r")"
    r")"
)
DefaultKnownWords = set()
KnownWords = set()


def init_known_words(DefaultKnownWords, KnownWords):
    default_dict = {
        unicodedata.normalize('NFC', word) for word in
        SpellChecker(case_sensitive=True).word_frequency
    }

    custom_dict = set()
    dict_path = os.path.join(os.path.dirname(__file__), 'dictionary.txt')
    with open(dict_path, 'r', encoding='utf-8') as fp:
        for line in fp:
            line = line.rstrip('\n').rstrip('\r')
            custom_dict.add(unicodedata.normalize('NFC', line))

    for dictionary, knownwords in [
            (default_dict, DefaultKnownWords), (custom_dict, KnownWords)]:
        for word in dictionary:
            # For words in full lower case, allow capitalization and all caps
            # (e.g. word, Word, & WORD); for other words, allow the word itself
            # and all caps (e.g. George & GEORGE or pH & PH)
            knownwords.add(word)
            if word == word.lower():
                knownwords.add(word.capitalize())
            knownwords.add(word.upper())

    KnownWords |= DefaultKnownWords


if not KnownWords:
    init_known_words(DefaultKnownWords, KnownWords)


def unit_test__tokenizer():
    text_and_results = [
        ("I'm a test", ["I'm", "test"]),
        ("Bahá'í", ["Bahá'í"]),  # containing composed character á
        ("Bahá'í", ["Bahá'í"]),  # containing decomposed sequence á
        ("This is Megingjörð", ["This", "is", "Megingjörð"]),
        ("Test 测试", ["Test"]),
        ("\"This'll do!\"", ["This'll", "do"]),
        ("Someone's test case", ["Someone", "test", "case"]),
        ("SOMEONE'S VERY IMPORTANT TEST",
            ["SOMEONE", "VERY", "IMPORTANT", "TEST"]),
        ("tic-tac-toe", ["tic", "tac", "toe"]),
        ("Ünicodé", ["Ünicodé"]),
        ("This apostrophe’s weird", ["This", "apostrophe’s", "weird"]),
        ("This apostrophe´s weird", ["This", "apostrophe´s", "weird"]),
        ("Foo i18n l10n bar", ["Foo", "bar"]),
        ("Lorem 123 ipsum -- dolor ||| sit",
            ["Lorem", "ipsum", "dolor", "sit"]),
        ("Lorem", ["Lorem"]),
        ("Lorem ipsum dolor sit amet",
            ["Lorem", "ipsum", "dolor", "sit", "amet"]),
        ("'Lorem ipsum dolor sit amet, consectetur adipiscing elit'",
            ["Lorem", "ipsum", "dolor", "sit", "amet", "consectetur",
                "adipiscing", "elit"]),
        ("Lorem, ipsum.  dolor?  sit!  amet.",
            ["Lorem", "ipsum", "dolor", "sit", "amet"]),
        ("Lorem ipsum, 'dolor?'  sit!?", ["Lorem", "ipsum", "dolor", "sit"]),
        ("Lorem: ipsum; dolor…", ["Lorem", "ipsum", "dolor"]),
        ('"Lorem"', ["Lorem"]),
        ("Lorem-ipsum", ["Lorem", "ipsum"]),
        ("«Lorem ipsum»", ["Lorem", "ipsum"]),  # used in some NPC dialogue
        ("*Lorem ipsum*", ["Lorem", "ipsum"]),
        ("&Lorem ipsum", ["Lorem", "ipsum"]),  # used in some NPC dialogue
        ("…Lorem", ["Lorem"]),
        ("'…Lorem ipsum…'", ["Lorem", "ipsum"]),
        ("Lorem <ipsum> <color_yellow>dolor</color>", ["Lorem", "dolor"]),
        ("<global_val:<u_val:foobar>>", []),
        ("Lorem (ipsum) [dolor sit] {amet}",
            ["Lorem", "ipsum", "dolor", "sit", "amet"]),
        ("Lorem ipsum/dolor/sit amet",
            ["Lorem", "ipsum", "dolor", "sit", "amet"]),
        ("Lorem.ipsum.dolor", []),
        ("Lorem %1$s ipsum %2$d dolor!", ["Lorem", "ipsum", "dolor"])
    ]

    fail = False
    for text, expected in text_and_results:
        result = list(Tokenizer.findall(text))
        if result != expected:
            print(f"test case: {text}\n"
                  f"expected: {expected}\nresult: {result}\n")
            fail = True

    if fail:
        raise RuntimeError("tokenizer test failed")



#!/usr/bin/env python3

# Use the Unicode Common Locale Data Repository to create code that
# creates ImGui glyph ranges for each locale, so that every locale can
# have properly rendered text.

import itertools
import sys
from cldr_language_helpers import alphabets


def add_preprocessor():
    print("#if defined(__GNUC__) or defined(__clang__)\n"
          "#define NOINLINE __attribute__ ((noinline))\n"
          "#else\n"
          "#define NOINLINE __declspec(noinline)\n"
          "#endif\n"
          "#if defined(__clang__)\n"
          "#define NOUNROLL _Pragma(\"clang loop unroll(disable)\")\n"
          "#elif defined(__GNUC__)\n"
          "#define NOUNROLL _Pragma(\"GCC unroll 0\")\n"
          "#else\n"
          "#define NOUNROLL\n"
          "#endif\n")


def main():
    print("// generated by get_translation_characters.py; example:\n"
          "//   ./build-scripts/get_translation_characters.py en ar cs da de "
          "el es fr hu id is it ja ko nb nl pl pt ru sr tr uk_UA zh_Hans "
          "zh_Hant > src/cldr/imgui-glyph-ranges.cpp\n\n"
          "// NOLINTBEGIN(cata-static-declarations,readability-function-size,"
          "modernize-avoid-c-arrays)\n")
    add_preprocessor()
    print("namespace {\n"
          "NOINLINE void AddGlyphs( ImFontGlyphRangesBuilder *b, "
          "ImWchar const *glyphp, ImWchar const *end) {\n"
          "  NOUNROLL\n"
          "  for( ; glyphp != end; ++glyphp ) {\n"
          "    b->AddChar(*glyphp);\n"
          "  }\n"
          "}\n"
          "} // namespace")
    try:
        for language in sys.argv[1:]:
            print_func(language)
        print("// NOLINTEND(cata-static-declarations,"
              "readability-function-size,modernize-avoid-c-arrays)\n")
        return 0
    except KeyError as x:
        print(f"Unknown language code “{x.args[0]}”", file=sys.stderr)
        return 1


def chunks(xs, n):
    n = max(1, n)
    return (xs[i:i + n] for i in range(0, len(xs), n))


def print_func(language):
    print(f"static void AddGlyphRangesFromCLDRFor{language.upper()}("
          "ImFontGlyphRangesBuilder *b) {")
    # All of the glyphs used this language
    chars = []
    for c in alphabets.ALPHABETS_BY_LANG_MAP[language]:
        for g in c:
            chars.append(ord(g))
    for c in alphabets.ALPHABETS_BY_LANG_MAP[language]:
        for g in c.upper():
            chars.append(ord(g))
    for c in alphabets.NUMBERS_BY_LANG_MAP[language]:
        for g in c:
            chars.append(ord(g))
    for c in alphabets.PUNCTUATION_BY_LANG_MAP[language]:
        for g in c:
            chars.append(ord(g))
    # Sort and remove duplicates, so we can detect sequences
    chars = [hex(c) for c in sorted(list(set(chars)))]

    print("  static constexpr ImWchar glyphs[] = {")
    char_chunks = list(chunks(chars, 16))
    for cs in itertools.islice(char_chunks, len(char_chunks) - 1):
        print(", ".join(cs) + ",")
    print(", ".join(char_chunks[-1]))
    print("  };")
    print("  AddGlyphs(b, glyphs, glyphs + std::extent_v<decltype(glyphs)>);")
    print("}")


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

import argparse
import json
import os

args = argparse.ArgumentParser()
args.add_argument("dir", action="store", help="specify data directory")
args_dict = vars(args.parse_args())

types = [
    "AMMO",
    "ARMOR",
    "BATTERY",
    "BIONIC_ITEM",
    "BOOK",
    "COMESTIBLE",
    "CONTAINER",
    "ENGINE",
    "GENERIC",
    "GUN",
    "GUNMOD",
    "MAGAZINE",
    "PET_ARMOR",
    "TOOL",
    "TOOLMOD",
    "TOOL_ARMOR",
]


def adjust_pockets(jo):
    if "pocket_data" not in jo:
        return None

    changed = False
    for pocket in jo["pocket_data"]:
        if "flag_restriction" in pocket:
            if ("BELT_CLIP" in pocket["flag_restriction"] or
                    "HELMET_HEAD_ATTACHMENT" in pocket["flag_restriction"]):
                pocket["transparent"] = True
                changed = True
                continue
        ripoff = 0
        encumb = 0
        if "ripoff" in pocket:
            ripoff = pocket["ripoff"]
        if "extra_encumbrance" in pocket:
            encumb = pocket["extra_encumbrance"]

        if ripoff > 0 and encumb > 0:
            pocket["transparent"] = True
            changed = True

    return jo if changed else None


def gen_new(path):
    change = False
    new_datas = list()
    # Having troubles with this script halting?
    # Uncomment below to find the file it's halting on
    #print(path)
    with open(path, "r", encoding="utf-8") as json_file:
        try:
            json_data = json.load(json_file)
        except json.JSONDecodeError:
            print(f"Json Decode Error at {path}")
            print("Ensure that the file is a JSON file consisting of"
                  " an array of objects!")
            return None

        for jo in json_data:
            if type(jo) is not dict:
                print(f"Incorrectly formatted JSON data at {path}")
                print("Ensure that the file is a JSON file consisting"
                      " of an array of objects!")
                break

            if "type" not in jo:
                continue

            if jo["type"] not in types:
                continue

            new_data = adjust_pockets(jo)

            if new_data is not None:
                jo = new_data
                change = True

    if len(new_datas) > 0:
        json_data.extend(new_datas)
    return json_data if change else None


def format_json(path):
    file_path = os.path.dirname(__file__)
    format_path_linux = os.path.join(file_path, "../format/json_formatter.cgi")
    path_win = "../format/JsonFormatter-vcpkg-static-Release-x64.exe"
    format_path_win = os.path.join(file_path, path_win)
    if os.path.exists(format_path_linux):
        os.system(f"{format_path_linux} {path}")
    elif os.path.exists(format_path_win):
        os.system(f"{format_path_win} {path}")
    else:
        print("No json formatter found")


# adjust items
json_path = os.path.join(args_dict["dir"], "json")
for root, directories, filenames in os.walk(json_path):
    for filename in filenames:
        if not filename.endswith(".json"):
            continue

        path = os.path.join(root, filename)
        new = gen_new(path)
        if new is not None:
            with open(path, "w", encoding="utf-8") as jf:
                json.dump(new, jf, ensure_ascii=False)
            format_json(path)

for f in os.scandir(os.path.join(args_dict["dir"], "mods")):
    if f.is_dir():
        for root, directories, filenames in os.walk(f.path):
            for filename in filenames:
                if not filename.endswith(".json"):
                    continue

                path = os.path.join(root, filename)
                new = gen_new(path)
                if new is not None:
                    with open(path, "w", encoding="utf-8") as jf:
                        json.dump(new, jf, ensure_ascii=False)
                    format_json(path)



#!/usr/bin/env python3
"Strip line numbers from comments in input .pot or .po file."

from __future__ import print_function

import sys
import os

usage = "USAGE: strip_line_numbers.py <filename> [<filename> ...]"


def strip_pot_file(filename):
    # Check pot file
    if not os.path.isfile(filename):
        print("Invalid filename: %s" % filename)
        sys.exit(1)
    try:
        to_write = open(filename, 'r', encoding="utf-8").readlines()
    except IOError as read_exc:
        print(read_exc)
        sys.exit(1)
    assert len(to_write) > 1  # Wrong .pot file

    to_write = strip_line_numbers(to_write)
    to_write = strip_repeated_comments(to_write)

    # Write the file back out with the line numbers stripped
    try:
        open(filename, 'w', encoding="utf-8").writelines(to_write)
    except IOError as write_exc:
        print(write_exc)
        sys.exit(1)


def strip_repeated_comments(lines):
    to_remove = []
    ln_length = len(lines)
    for i in range(ln_length):
        line = lines[i]
        if i < (ln_length - 1):
            next_line = lines[i + 1]
        else:
            next_line = None
        # only act on line-comments
        if not line.startswith("#:"):
            continue
        if next_line is not None and line == next_line:
            to_remove.append(i)
            continue
    # Remove obsolete strings from list
    to_remove.reverse()
    for line_num in to_remove:
        del lines[line_num]

    return lines


def strip_line_numbers(lines):
    "Strip line numbers from the specified list of strings."

    # strip line-numbers
    for i in range(len(lines)):
        line = lines[i]
        # only act on line-comments
        if not line.startswith("#:"):
            continue
        # line comments are of the format: "#: <file>:<line> <file>:<line>"
        items = line.split()
        newitems = ["#:"]
        # replace line elements while yanking the line number,
        # and checking for duplicates
        for item in items[1:]:
            newname = item.split(':')[0]
            if newname not in newitems:
                newitems.append(newname)
        # replace the old line with the new
        newline = ' '.join(newitems) + '\n'
        lines[i] = newline

    return lines


if __name__ == "__main__":
    # get filename(s) from sys.argv and strip
    if len(sys.argv) < 2:
        # ideally this would read from stdin and input to stdout,
        # but not need to do that yet.
        print(usage)
        sys.exit(1)

    for filename in sys.argv[1:]:
        strip_pot_file(filename)



from ..helper import get_singular_name
from ..write_text import write_text


def parse_movement_mode(json, origin):
    name = get_singular_name(json["name"])
    write_text(json["name"], origin, comment="Movement mode name")
    write_text(json["character"], origin,
               comment="Character displayed in the move menu for "
               "movement mode \"{}\"".format(name))
    write_text(json["panel_char"], origin,
               comment="Character displayed in the panel for "
               "movement mode \"{}\"".format(name))
    write_text(json["change_good_none"], origin,
               comment="Successfully switched to movement mode "
               "\"{}\" with no steed".format(name))
    write_text(json["change_good_animal"], origin,
               comment="Successfully switched to movement mode "
               "\"{}\" with animal steed".format(name))
    write_text(json["change_good_mech"], origin,
               comment="Successfully switched to movement mode "
               "\"{}\" with mechanical steed".format(name))
    if "change_bad_none" in json:
        write_text(json["change_bad_none"], origin,
                   comment="Failed to switched to movement mode "
                   "\"{}\" with no steed".format(name))
    if "change_bad_animal" in json:
        write_text(json["change_bad_animal"], origin,
                   comment="Failed to switched to movement mode "
                   "\"{}\" with animal steed".format(name))
    if "change_bad_mech" in json:
        write_text(json["change_bad_mech"], origin,
                   comment="Failed to switched to movement mode "
                   "\"{}\" with mechanical steed".format(name))



#!/usr/bin/env python3

import argparse
import json
import os

args = argparse.ArgumentParser()
args.add_argument("dir", action="store", help="specify data directory")
args_dict = vars(args.parse_args())

types = [
    "AMMO",
    "ARMOR",
    "BATTERY",
    "BIONIC_ITEM",
    "BOOK",
    "COMESTIBLE",
    "CONTAINER",
    "ENGINE",
    "GENERIC",
    "GUN",
    "GUNMOD",
    "MAGAZINE",
    "PET_ARMOR",
    "TOOL",
    "TOOLMOD",
    "TOOL_ARMOR",
]

oversized = list()
changed = dict()
affected = dict()
items = dict()
recipes = dict()
group_migration = dict()


def get_id(jo):
    if "id" in jo:
        return jo["id"]
    elif "abstract" in jo:
        return jo["abstract"]


def get_parent_value(id, property, src):
    if id not in items[src]:
        if src != "dda":
            return get_parent_value(id, property, "dda")
        else:
            return None
    if property not in items[src][id]:
        if "copy-from" in items[src][id]:
            if id == items[src][id]["copy-from"]:
                if src == "dda":
                    return None
                return get_parent_value(id, property, "dda")
            return get_parent_value(items[src][id]["copy-from"], property, src)
        else:
            return None
    return items[src][id][property]


def get_value(jo, property, src):
    if property in jo:
        return jo[property]
    if "copy-from" in jo:
        return get_parent_value(jo["copy-from"], property, src)
    return None


def get_jo(id, src):
    if src not in items:
        if src != "dda":
            return get_jo(id, "dda")
        return None
    if id not in items[src]:
        if src != "dda":
            return get_jo(id, "dda")
        return None
    return items[src][id]


def remove_charges(jo, src):
    phase = get_value(jo, "phase", src)
    if phase is not None and phase != "solid":
        return

    if "stackable" in jo and jo["stackable"]:
        return

    if src not in affected:
        affected[src] = dict()
    copy_from = None
    if "copy-from" in jo:
        copy_from = jo["copy-from"]
    affected[src][get_id(jo)] = copy_from

    has_volume = "volume" in jo
    has_charges = "charges" in jo or "stack_size" in jo

    if not has_volume and not has_charges:
        return

    charges = get_value(jo, "charges", src)
    stack_size = get_value(jo, "stack_size", src)
    if stack_size is None:
        if charges is None:
            return
        stack_size = charges
    if charges is None:
        charges = -1

    vol = get_value(jo, "volume", src)
    vol_arr = vol.split()
    if len(vol_arr) != 2:
        pos = vol.find("ml")
        unit = "ml"
        if pos == -1:
            pos = vol.find("L")
            unit = "L"
        vol = vol[0:pos]
    else:
        vol = vol_arr[0]
        unit = vol_arr[1]
    vol = int(vol)
    if unit == "L":
        vol *= 1000
        unit = "ml"

    if vol / stack_size < 1:
        oversized.append(get_id(jo))

    vol = max(1, vol // stack_size)

    if (vol % 1000 == 0):
        vol = vol // 1000
        unit = "L"

    jo["volume"] = str(vol) + " " + unit
    if "charges" in jo:
        jo.pop("charges")
    if "stack_size" in jo:
        jo.pop("stack_size")
    if src not in changed:
        changed[src] = dict()
    changed[src][get_id(jo)] = [charges, stack_size]

    return jo


def multiply_values(lhs, rhs):
    left_list = type(lhs) is list
    right_list = type(rhs) is list
    if left_list or right_list:
        ret = list()
        val_min = 0
        val_max = 0
        if left_list:
            val_min = lhs[0]
            val_max = lhs[1]
        else:
            val_min = lhs
            val_max = lhs
        if right_list:
            val_min *= rhs[0]
            val_max *= rhs[1]
        else:
            val_min *= rhs
            val_max *= rhs
        ret.append(val_min)
        ret.append(val_max)
    else:
        ret = lhs * rhs

    return ret


def get_min(entry, property):
    min = -1

    if property in entry:
        min = entry[property]
        if type(min) is list:
            min = min[0]

    if property + "-min" in entry:
        min = entry.pop(property + "-min")

    return min


def get_max(entry, property):
    max = -1

    if property in entry:
        max = entry.pop(property)
        if type(max) is list:
            max = max[1]

    if property + "-max" in entry:
        max = entry.pop(property + "-max")

    return max


def get_changed(id, src):
    if src not in changed:
        if src != "dda":
            return get_changed(id, "dda")
        return None

    if id in changed[src]:
        return changed[src][id]

    if id in changed["dda"]:
        return changed["dda"][id]

    return None


def get_maybe_changed(id, src):
    changed = get_changed(id, src)
    if changed is not None:
        return changed

    if src not in affected:
        if src != "dda":
            return get_maybe_changed(id, "dda")
        return None

    if id in affected[src] and affected[src][id] is not None:
        return get_affected_parent(affected[src][id], src)

    return None


def get_affected_parent(copy_from, src):
    if src in changed and copy_from in changed[src]:
        return changed[src][copy_from]

    if src not in items or copy_from not in items[src]:
        if src == "dda":
            return None
        return get_affected_parent(copy_from, "dda")

    if "copy-from" not in items[src][copy_from]:
        return None

    if copy_from == items[src][copy_from]["copy-from"]:
        if src == "dda":
            return None
        return get_affected_parent(copy_from, "dda")

    return get_affected_parent(items[src][copy_from]["copy-from"], src)


def make_group(entry, default_container):
    jo = dict()
    jo["type"] = "item_group"
    new_entry = dict()
    new_entry["item"] = entry.pop("item")
    if default_container is not None:
        new_entry["container-item"] = "null"
    id_count_min = 0
    id_count_max = 0
    group_count = ("charges" in entry and type(entry["charges"]) is list and
                   entry["charges"][0] == 0 and entry["charges"][1] == 1)
    if group_count:
        entry["count"] = entry.pop("charges")
    if "charges" in entry:
        if type(entry["charges"]) is not list and entry["charges"] <= 1:
            entry.pop("charges")
        else:
            new_entry["count"] = entry.pop("charges")
            if type(new_entry["count"]) is list:
                id_count_min = new_entry["count"][0]
                id_count_max = new_entry["count"][1]
            else:
                id_count_min = new_entry["count"]
                id_count_max = new_entry["count"]
    if "charges-min" in entry:
        new_entry["count-min"] = entry.pop("charges-min")
        id_count_min = new_entry["count-min"]
    if "charges-max" in entry:
        new_entry["count-max"] = entry.pop("charges-max")
        id_count_max = new_entry["count-max"]
    jo["id"] = new_entry["item"] + "_" + entry["container-item"] + "_"
    if id_count_max == id_count_min:
        if id_count_max <= 0:
            id_count_max = 1
        jo["id"] += str(id_count_max)
    elif id_count_max <= 0:
        jo["id"] += str(id_count_min) + "_inf"
    else:
        jo["id"] += str(id_count_min) + "_" + str(id_count_max)
    jo["subtype"] = "collection"
    jo["//"] = "This group was created automatically and may contain errors."
    jo["container-item"] = entry.pop("container-item")
    jo["entries"] = [new_entry]
    return jo


def has_group(id, src):
    if src not in group_migration:
        if src != "dda":
            return has_group(id, "dda")
        return False
    if id not in group_migration[src]:
        if src != "dda":
            return has_group(id, "dda")
        return False
    return True


def adjust_item_group(jo, src):
    item_array = []
    if "entries" in jo:
        item_array = jo["entries"]
    elif "items" in jo:
        item_array = jo["items"]
    else:
        return

    ret = [jo]
    change = False
    for entry in item_array:
        if "item" not in entry:
            continue

        use_src = src
        if src not in affected:
            use_src = "dda"
        if entry["item"] not in affected[use_src]:
            continue

        has_charges = ("charges" in entry or "charges-min" in entry or
                       "charges-max" in entry)

        base_item = get_jo(entry["item"], src)
        if not has_charges:
            base_charges = get_value(base_item, "charges", src)
            if base_charges is not None and base_charges > 1:
                entry["charges"] = base_charges
            else:
                continue

        base_item = get_jo(entry["item"], src)
        default_container = get_value(base_item, "container", src)
        if ("container-item" not in entry and default_container is not None and
                default_container != "null"):
            entry["container-item"] = default_container

        if "container-item" in entry:
            new_group = make_group(entry, default_container)
            entry["group"] = new_group["id"]
            if not has_group(new_group["id"], src):
                if src not in group_migration:
                    group_migration[src] = list()
                group_migration[src].append(new_group["id"])
                ret.append(new_group)
            change = True
            continue

        change = True
        charge_arr = "charge" in entry and type(entry["charge"]) is list
        make_min = ("count-min" in entry or "charges-min" in entry or
                    ("count-max" in entry and charge_arr))
        make_max = ("count-max" in entry or "charges-max" in entry or
                    ("count-min" in entry and charge_arr))

        changes = get_maybe_changed(entry["item"], src)
        default = 1
        if changes is not None:
            default = changes[0]

        if make_min:
            # get_min must be called before get_max
            count_min = get_min(entry, "count")
            charges_min = get_min(entry, "charges")
            if charges_min < 0 and default > 1:
                charges_min = default
            if count_min < 0:
                if charges_min >= 0:
                    entry["count-min"] = charges_min
            else:
                if charges_min >= 0:
                    entry["count-min"] = count_min * charges_min
                else:
                    entry["count-min"] = count_min
        if make_max:
            count_max = get_max(entry, "count")
            charges_max = get_max(entry, "charges")
            if charges_max < 0 and default > 1:
                charges_max = default
            if count_max < 0:
                if charges_max >= 0:
                    entry["count-max"] = charges_max
            else:
                if charges_max >= 0:
                    entry["count-max"] = count_max * charges_max
                else:
                    entry["count-max"] = count_max
        if not make_min and not make_max:
            if "count" in entry:
                if "charges" in entry:
                    entry["count"] = multiply_values(entry["count"],
                                                     entry.pop("charges"))
                elif default > 1:
                    entry["count"] = multiply_values(entry["count"], default)
                else:
                    change = False
            else:
                if "charges" in entry:
                    charges = entry.pop("charges")
                    if type(charges) is list or charges > 1:
                        entry["count"] = charges
                elif default > 1:
                    entry["count"] = default
                else:
                    change = False

    if change:
        return ret


def get_parent_charges(result, src):
    if src not in recipes:
        if src != "dda":
            get_parent_charges(result, "dda")
        return None

    if result not in recipes[src]:
        if src != "dda":
            get_parent_charges(result, "dda")
        return None

    if "charges" in recipes[src][result]:
        return recipes[src][result]["charges"]

    if "copy-from" in recipes[src][result]:
        return get_parent_charges(recipes[src][result]["copy-from"], src)

    return None


def adjust_recipe(jo, src):
    if "charges" in jo or "result" not in jo:
        return

    aff_it = get_maybe_changed(jo["result"], src)

    if aff_it is None or aff_it[0] <= 1:
        return

    if "copy-from" in jo:
        charges = get_parent_charges(jo["copy-from"], src)
        if charges is not None:
            return

    jo["charges"] = aff_it[0]
    return jo


def gen_new(path, src, is_item=True):
    change = False
    new_datas = list()
    # Having troubles with this script halting?
    # Uncomment below to find the file it's halting on
    #print(path)
    with open(path, "r", encoding="utf-8") as json_file:
        try:
            json_data = json.load(json_file)
        except json.JSONDecodeError:
            print(f"Json Decode Error at {path}")
            print("Ensure that the file is a JSON file consisting of"
                  " an array of objects!")
            return None

        for jo in json_data:
            if type(jo) is not dict:
                print(f"Incorrectly formatted JSON data at {path}")
                print("Ensure that the file is a JSON file consisting"
                      " of an array of objects!")
                break

            if "type" not in jo:
                continue

            new_data = None
            if is_item and jo["type"] == "COMESTIBLE":
                new_data = remove_charges(jo, src)
            if not is_item and jo["type"] == "item_group":
                new_groups = adjust_item_group(jo, src)
                if type(new_groups) is list and len(new_groups) > 0:
                    new_data = new_groups[0]
                    if len(new_groups) > 1:
                        new_datas.extend(new_groups[1:])
                else:
                    new_data = new_groups
            if not is_item and jo["type"] == "recipe":
                new_data = adjust_recipe(jo, src)

            if new_data is not None:
                jo = new_data
                change = True

    if len(new_datas) > 0:
        json_data.extend(new_datas)
    return json_data if change else None


def format_json(path):
    file_path = os.path.dirname(__file__)
    format_path_linux = os.path.join(file_path, "../format/json_formatter.cgi")
    path_win = "../format/JsonFormatter-vcpkg-static-Release-x64.exe"
    format_path_win = os.path.join(file_path, path_win)
    if os.path.exists(format_path_linux):
        os.system(f"{format_path_linux} {path}")
    elif os.path.exists(format_path_win):
        os.system(f"{format_path_win} {path}")
    else:
        print("No json formatter found")


def collect_data(path, src):
    with open(path, "r", encoding="utf-8") as json_file:
        try:
            json_data = json.load(json_file)
        except json.JSONDecodeError:
            print(f"Json Decode Error at {path}")
            print("Ensure that the file is a JSON file consisting of"
                  " an array of objects!")
            return

        for jo in json_data:
            if type(jo) is not dict:
                print(f"Incorrectly formatted JSON data at {path}")
                print("Ensure that the file is a JSON file consisting"
                      " of an array of objects!")
                break

            if "type" not in jo:
                continue

            if jo["type"] in types:
                item = dict()
                if "volume" in jo:
                    item["volume"] = jo["volume"]
                if "charges" in jo:
                    item["charges"] = jo["charges"]
                if "stack_size" in jo:
                    item["stack_size"] = jo["stack_size"]
                if "copy-from" in jo:
                    item["copy-from"] = jo["copy-from"]
                if "container" in jo:
                    item["container"] = jo["container"]
                if "phase" in jo:
                    item["phase"] = jo["phase"]

                if src not in items:
                    items[src] = dict()
                items[src][get_id(jo)] = item

            if jo["type"] == "recipe" and "result" in jo:
                recipe = dict()
                if "charges" in jo:
                    recipe["charges"] = jo["charges"]
                if "copy-from" in jo:
                    recipe["copy-from"] = jo["copy-from"]

                if src not in recipes:
                    recipes[src] = dict()
                recipes[src][jo["result"]] = recipe


# collect item data with source
json_path = os.path.join(args_dict["dir"], "json")
for root, directories, filenames in os.walk(json_path):
    for filename in filenames:
        if not filename.endswith(".json"):
            continue

        path = os.path.join(root, filename)
        collect_data(path, "dda")

for f in os.scandir(os.path.join(args_dict["dir"], "mods")):
    if f.is_dir():
        for root, directories, filenames in os.walk(f.path):
            for filename in filenames:
                if not filename.endswith(".json"):
                    continue

                path = os.path.join(root, filename)
                collect_data(path, f.name)

# adjust items
for root, directories, filenames in os.walk(json_path):
    for filename in filenames:
        if not filename.endswith(".json"):
            continue

        path = os.path.join(root, filename)
        new = gen_new(path, "dda")
        if new is not None:
            with open(path, "w", encoding="utf-8") as jf:
                json.dump(new, jf, ensure_ascii=False)
            format_json(path)

for f in os.scandir(os.path.join(args_dict["dir"], "mods")):
    if f.is_dir():
        for root, directories, filenames in os.walk(f.path):
            for filename in filenames:
                if not filename.endswith(".json"):
                    continue

                path = os.path.join(root, filename)
                new = gen_new(path, f.name)
                if new is not None:
                    with open(path, "w", encoding="utf-8") as jf:
                        json.dump(new, jf, ensure_ascii=False)
                    format_json(path)

# adjust itemgroups
for root, directories, filenames in os.walk(json_path):
    for filename in filenames:
        if not filename.endswith(".json"):
            continue

        path = os.path.join(root, filename)
        new = gen_new(path, "dda", False)
        if new is not None:
            with open(path, "w", encoding="utf-8") as jf:
                json.dump(new, jf, ensure_ascii=False)
            format_json(path)

for f in os.scandir(os.path.join(args_dict["dir"], "mods")):
    if f.is_dir():
        for root, directories, filenames in os.walk(f.path):
            for filename in filenames:
                if not filename.endswith(".json"):
                    continue

                path = os.path.join(root, filename)
                new = gen_new(path, f.name, False)
                if new is not None:
                    with open(path, "w", encoding="utf-8") as jf:
                        json.dump(new, jf, ensure_ascii=False)
                    format_json(path)

with open('oversized_charge_removal.txt', 'w') as outfile:
    outfile.write('\n'.join(id for id in oversized))

with open('removed_charges.txt', 'w') as outfile:
    header = 'id | previous default charges | previous stack_size | src\n'
    outfile.write(header)
    for src, it in changed.items():
        outfile.write('\n'.join(f'{k} {v[0]} {v[1]} {src}' for k, v in
                                it.items()) + '\n')

with open('itemgroup_migration.txt', 'w') as outfile:
    header = 'new group id | src\n'
    outfile.write(header)
    for src, groups in group_migration.items():
        outfile.write('\n'.join(f'{group} {src}' for group in groups) + '\n')



from ..helper import get_singular_name
from ..write_text import write_text


def parse_option_slider(json, origin):
    name = get_singular_name(json["name"])
    write_text(json["name"], origin,
               comment="Name of an option slider", c_format=False)
    if "levels" in json:
        for level in json["levels"]:
            if "name" in level:
                write_text(level["name"], origin,
                           comment="Name of a slider position in the "
                           "\"{}\" option slider".format(name))
            if "description" in level:
                write_text(level["description"], origin,
                           comment="Description of a slider position in the "
                           "\"{}\" option slider".format(name))



#!/usr/bin/env python3

import getopt
import glob
import json
import os
import re
import subprocess
import sys


def print_help():
    print(
        "\n"
        "Update faction camp blueprints with autocalculated requirements\n"
        "from unit test log.\n"
        "This tool requires tools/format/json_formatter.\n"
        "\n"
        "  --help   prints this message\n"
        "  --logfile=<logfile>\n"
        "           specify the path to unit test log file\n"
        "  --action=<value>\n"
        "           what to do with reported inconsistencies. (optional)\n"
        "             update: update with suggested requirements (default)\n"
        "             suppress: suppress inconsistency warnings\n")


def main(argv):
    try:
        opts, args = getopt.getopt(argv, "", ["help", "logfile=", "action="])
    except getopt.GetoptError:
        print_help()
        return

    test_log = None
    suppress = False
    for opt, arg in opts:
        if opt == "--help":
            print_help()
            return
        if opt == "--logfile":
            test_log = arg
        if opt == "--action":
            if arg == "update":
                suppress = False
            elif arg == "suppress":
                suppress = True
            else:
                print_help()
                return
    if not test_log:
        print_help()
        return

    json_dirs = {
        "data/core",
        "data/json",
        "data/mods",
    }

    auto_update_blueprint = re.compile("~~~ auto-update-blueprint: (.+)")
    auto_update_blueprint_end = re.compile("~~~ end-auto-update")
    json_filename = re.compile("[^.].*\\.json")

    update_blueprints = dict()

    with open(test_log, 'r', encoding='utf-8') as fs:
        while True:
            line = fs.readline()
            if not line:
                break
            match_result = auto_update_blueprint.match(line)
            if match_result:
                ident = match_result.group(1)
                reqs = ""
                complete = False
                while True:
                    line = fs.readline()
                    if not line:
                        complete = False
                        break
                    if auto_update_blueprint_end.match(line):
                        complete = True
                        break
                    else:
                        reqs += line
                if complete:
                    update_blueprints[ident] = json.loads(reqs)
                    print("{} needs updating".format(ident))

    if len(update_blueprints) == 0:
        print("no inconsistency reported in the test log")
        return

    for json_dir in json_dirs:
        print("walking dir {}".format(json_dir))
        for root, dirs, files in os.walk(json_dir):
            for file in files:
                json_path = os.path.join(root, file)
                content = None
                changed = False
                if json_filename.match(file):
                    try:
                        with open(json_path, 'r', encoding='utf-8') as fs:
                            content = json.load(fs)
                    except Exception:
                        sys.stderr.write('Error parsing %r\n' % json_path)
                        raise
                if type(content) is list:
                    for obj in content:
                        if not (type(obj) is dict and
                                "type" in obj and obj["type"] == "recipe" and
                                ("result" in obj or "abstract" in obj)):
                            continue
                        ident = None
                        if "abstract" in obj:
                            ident = obj["abstract"]
                        else:
                            ident = obj["result"]
                            if "id_suffix" in obj:
                                ident += "_" + obj["id_suffix"]
                        if ident in update_blueprints:
                            if suppress:
                                obj["check_blueprint_needs"] = False
                            else:
                                obj["blueprint_needs"] = \
                                    update_blueprints[ident]
                            if not changed:
                                changed = True
                                print("updating {}".format(json_path))
                if changed:
                    with open(json_path, 'w', encoding='utf-8') as fs:
                        json.dump(content, fs, indent=2)
                    json_formatter_name = glob.glob(
                        'tools/format/json_formatter.[ec]*')
                    assert len(json_formatter_name) == 1
                    subprocess.run([json_formatter_name[0], json_path],
                                   stdout=subprocess.DEVNULL)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
"""Update a tags file with locations of the definitions of CDDA json entities.

If you already have a tags file with some data in, this will only
replace tags in json files, not e.g. cpp files, so it should be safe
to use after running e.g. ctags.
"""

import argparse
import json
import os
import sys

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))
TOP_DIR = os.path.normpath(os.path.join(SCRIPT_DIR, "..", ".."))
JSON_DIR = os.path.join(TOP_DIR, "data")
TAGS_FILE = os.path.join(TOP_DIR, "tags")


def cdda_style_json(v):
    if type(v) is str:
        return json.dumps(v)
    elif type(v) is list:
        return f'[ {", ".join(cdda_style_json(e) for e in v)} ]'
    else:
        raise RuntimeError('Unexpected type')


def make_tags_line(id_key, id, full_id, filename):
    length_limit = 120
    pattern = f'/"{id_key}": {cdda_style_json(full_id)}/'
    if len(pattern) > length_limit - 5:
        pattern = f'/"{id}"/'
    return '\t'.join((id, filename, pattern)).encode('utf-8')


def is_json_tag_line(line):
    return b'.json\t' in line


def main(args):
    parser = argparse.ArgumentParser(description=__doc__)

    parser.parse_args(args)

    definitions = []

    # JSON keys to generate tags for
    id_keys = (
        'id', 'abstract', 'ident',
        'nested_mapgen_id', 'om_terrain', 'update_mapgen_id', 'result')

    def add_definition(id_key, id, full_id, relative_path):
        if not id:
            return
        if type(id) is str:
            definitions.append((id_key, id, full_id, relative_path))
        elif type(id) is list:
            for i in id:
                add_definition(id_key, i, full_id, relative_path)

    for dirpath, dirnames, filenames in os.walk(JSON_DIR):
        for filename in filenames:
            if filename.endswith('.json'):
                full_path = os.path.join(dirpath, filename)
                assert full_path.startswith(TOP_DIR)
                relative_path = full_path[len(TOP_DIR):].lstrip(os.path.sep)
                with open(full_path, encoding='utf-8') as file:
                    try:
                        json_data = json.load(file)
                    except Exception as err:
                        sys.stderr.write(
                            "Problem reading file %s, reason: %s" %
                            (filename, err))
                        continue
                    if type(json_data) is dict:
                        json_data = [json_data]
                    elif type(json_data) is not list:
                        sys.stderr.write(
                            "Problem parsing data from file %s, reason: "
                            "expected a list." % filename)
                        continue

                    # Check each object in json_data for taggable keys
                    for obj in json_data:
                        for id_key in id_keys:
                            if id_key in obj:
                                id = obj[id_key]
                                add_definition(id_key, id, id, relative_path)

    json_tags_lines = [make_tags_line(*d) for d in definitions]
    existing_tags_lines = []
    try:
        with open(TAGS_FILE, 'rb') as tags_file:
            existing_tags_lines = tags_file.readlines()
    except FileNotFoundError:
        pass

    existing_tags_lines = [
        line.rstrip(b'\n') for line in existing_tags_lines if
        not is_json_tag_line(line)]

    all_tags_lines = sorted(json_tags_lines + existing_tags_lines)

    with open(TAGS_FILE, 'wb') as tags_file:
        tags_file.write(b'\n'.join(all_tags_lines))


if __name__ == '__main__':
    main(sys.argv[1:])



from ..helper import get_singular_name
from ..write_text import write_text


def parse_spell(json, origin):
    id = ""
    name = ""

    if "id" in json:
        id = json["id"]
    else:
        id = json["abstract"]

    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin,
                   comment="Name of spell \"{}\"".format(id))

    if not name == "":
        id = id + " ( " + name + " )"

    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of spell \"{}\"".format(id))

    if "sound_description" in json:
        write_text(json["sound_description"], origin,
                   comment="Sound description of spell \"{}\"".format(id))

    if "message" in json:
        write_text(json["message"], origin,
                   comment="Message of spell \"{}\"".format(id))



from ..helper import get_singular_name
from ..write_text import write_text


def parse_skill(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Skill name")

    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of skill \"{}\"".format(name))

    for level in json.get("level_descriptions_theory", []):
        write_text(level["description"], origin,
                   comment="Level description of skill \"{}\"".format(name))

    for level in json.get("level_descriptions_practice", []):
        write_text(level["description"], origin,
                   comment="Level description of skill \"{}\"".format(name))



#!/usr/bin/env python3

import json
import argparse
import copy
import math
from operator import itemgetter


STRIDE_X = 3
STRIDE_Y = 2
MIN_X = 10000
MIN_Y = 10000
OM_SPEC_TYPES = ["overmap_special", "city_building"]
KEYED_TERMS = [
    "terrain", "furniture", "fields", "npcs", "signs", "vendingmachines",
    "toilets", "gaspumps", "items", "monsters", "vehicles", "item", "traps",
    "monster", "rubble", "liquids", "sealed_item", "graffiti", "mapping"
]
PLACE_TERMS = ["set", "place_groups"]
MAP_ROTATE = [
    {"dir": "_south", "x": 1, "y": 1},
    {"dir": "_north", "x": 0, "y": 0}
]


def x_y_bucket(x, y):
    return "{}__{}".format(math.floor((x - MIN_X) / STRIDE_X),
                           math.floor((y - MIN_Y) / STRIDE_Y))


def x_y_sub(x, y, is_north):
    if is_north:
        return "{}__{}".format((x - MIN_X) % STRIDE_X, (y - MIN_Y) % STRIDE_Y)
    else:
        return "{}__{}".format((x - MIN_X - 1) % STRIDE_X,
                               (y - MIN_Y - 1) % STRIDE_Y)


def x_y_simple(x, y):
    return "{}__{}".format(x, y)


def get_data(argsDict, resource_name):
    resource = []
    resource_sources = argsDict.get(resource_name, [])
    if not isinstance(resource_sources, list):
        resource_sources = [resource_sources]
    for resource_filename in resource_sources:
        if resource_filename.endswith(".json"):
            try:
                with open(
                        resource_filename, encoding="utf-8") as resource_file:
                    resource += json.load(resource_file)
            except FileNotFoundError:
                exit("Failed: could not find {}".format(resource_filename))
        else:
            print(("Invalid filename {}".format(resource_filename)))
    return resource


def adjacent_to_set(x, y, coord_set):
    for coords in coord_set:
        if y == coords["y"] and abs(x - coords["x"]) == 1:
            return True
        if x == coords["x"] and abs(y - coords["y"]) == 1:
            return True


def validate_keyed(key_term, old_obj, entry):
    new_keyset = entry["object"].get(term, {})
    old_keyset = old_obj.get(key_term, {})
    if new_keyset:
        for old_key, old_val in old_keyset.items():
            new_keyset.setdefault(old_key, old_val)
            if new_keyset[old_key] != old_val:
                return False
    else:
        new_keyset = old_keyset
    return new_keyset


# make sure that all keyed entries have the same key values and the maps have
# the same weight and fill_ter.  Don't try to resolve them.
def validate_old_map(old_map, entry):
    old_obj = old_map.get("object", {})

    if entry["weight"] and old_map.get("weight"):
        if entry["weight"] != old_map.get("weight"):
            return False
    if entry["object"].get("fill_ter") and old_obj.get("fill_ter") and \
            entry["object"]["fill_ter"] != old_obj.get("fill_ter"):
        return False

    new_palettes = entry.get("palettes", {})
    old_palettes = old_obj.get("palettes")
    if new_palettes:
        for palette in old_palettes:
            if palette not in new_palettes:
                return False
    else:
        new_palettes = old_palettes

    keysets = {}
    for key_term in KEYED_TERMS:
        new_keyset = validate_keyed(key_term, old_obj, entry)
        if new_keyset:
            keysets[key_term] = new_keyset
        elif new_keyset != {}:
            return False

    if not entry["weight"]:
        entry["weight"] = old_map.get("weight", 0)
    if not entry["object"].get("fill_ter"):
        entry["object"]["fill_ter"] = old_obj.get("fill_ter", "")
    for key_term, new_keyset in keysets.items():
        entry["object"][key_term] = new_keyset
    if new_palettes:
        entry["object"]["palettes"] = new_palettes

    return True


# adjust the X, Y co-ords of a place_ entry to match the new map
def adjust_place(term, old_obj, offset_x, offset_y):
    def adjust_coord(x_or_y, new_entry, old_entry, offset):
        val = old_entry.get(x_or_y, "False")
        if val == "False":
            return False
        if isinstance(val, list):
            val[0] += offset
            val[1] += offset
        else:
            val += offset
        new_entry[x_or_y] = val

    results = []
    for old_entry in old_obj.get(term, []):
        new_entry = copy.deepcopy(old_entry)
        if offset_x:
            adjust_coord("x", new_entry, old_entry, offset_x)
            adjust_coord("x2", new_entry, old_entry, offset_x)
        if offset_y:
            adjust_coord("y", new_entry, old_entry, offset_y)
            adjust_coord("y2", new_entry, old_entry, offset_y)
        results.append(new_entry)
    return results


args = argparse.ArgumentParser(
    description="Merge individual OMT maps into blocks of maps.")
args.add_argument("mapgen_sources", action="store", nargs="+",
                  help="specify jsons file to convert to blocks.")
args.add_argument("specials_sources", action="store", nargs="+",
                  help="specify json file with overmap special data.")
args.add_argument(
    "--x", dest="stride_x", action="store",
    help="number of horizontal maps in each block.  "
         "Defaults to {}.".format(STRIDE_X))
args.add_argument(
    "--y", dest="stride_y", action="store",
    help="number of vertictal maps in each block.  "
         "Defaults to {}.".format(STRIDE_Y))
args.add_argument("--output", dest="output_name", action="store",
                  help="Name of output file.  Defaults to the command line.")
argsDict = vars(args.parse_args())

mapgen = get_data(argsDict, "mapgen_sources")
specials = get_data(argsDict, "specials_sources")
string_x = argsDict.get("stride_x")
if string_x and int(string_x):
    STRIDE_X = int(string_x)
string_y = argsDict.get("stride_y")
if string_y and int(string_y):
    STRIDE_Y = int(string_y)

output_name = argsDict.get("output_name", "")
if output_name and not output_name.endswith(".json"):
    output_name += ".json"

# very first pass, sort the overmaps and find the minimum X and Y values
for special in specials:
    if special.get("type") in OM_SPEC_TYPES:
        overmaps = special.get("overmaps")
        if not overmaps:
            continue
        overmaps.sort(key=lambda om_data: om_data.get("point", [1000, 0])[0])
        MIN_X = overmaps[0].get("point", [1000, 0])[0]
        overmaps.sort(key=lambda om_data: om_data.get("point", [0, 1000])[1])
        MIN_Y = overmaps[0].get("point", [0, 1000])[1]

# create the merge sets of maps
merge_sets = {}
for special in specials:
    if special.get("type") in OM_SPEC_TYPES:
        overmaps = special.get("overmaps")
        for om_data in overmaps:
            om_map = om_data.get("overmap")
            for map_dir in MAP_ROTATE:
                if om_map.endswith(map_dir["dir"]):
                    om_map = om_map.split(map_dir["dir"])[0]
                    break
            om_point = om_data.get("point", [])
            if len(om_point) == 3:
                is_north = map_dir["dir"] == "_north"
                x = om_point[0]
                y = om_point[1]
                z = om_point[2]
                merge_sets.setdefault(z, {})
                merge_sets[z].setdefault(x_y_bucket(x, y), {})
                merge_sets[z][x_y_bucket(x, y)][x_y_sub(x, y, is_north)] = \
                    om_map

# convert the mapgen list into a dictionary for easier access
map_dict = {}
new_mapgen = []
for om_map in mapgen:
    if om_map.get("type") != "mapgen":
        new_mapgen.append(om_map)
        continue
    om_id = om_map["om_terrain"]
    if isinstance(om_id, list):
        if len(om_id) == 1:
            om_id = om_id[0]
        else:
            continue
    map_dict[om_id] = om_map

# dynamically expand the list of "place_" terms
for term in KEYED_TERMS:
    PLACE_TERMS.append("place_" + term)

basic_entry = {
    "method": "json",
    "object": {
        "fill_ter": "",
        "rows": [],
        "palettes": [],
    },
    "om_terrain": [],
    "type": "mapgen",
    "weight": 0
}

# debug: make sure the merge sets look correct
#print("mergesets: {}".format(json.dumps(merge_sets, indent=2)))

# finally start merging maps
for z, zlevel in merge_sets.items():
    for x_y, mapset in zlevel.items():
        # first, split the mergeset into chunks with common KEYED_TERMS using a
        # weird floodfill
        chunks = []
        chunks = [{"maps": [], "entry": copy.deepcopy(basic_entry)}]
        for y in range(0, STRIDE_Y):
            for x in range(0, STRIDE_X):
                om_id = mapset.get(x_y_simple(x, y), "")
                if not om_id:
                    continue
                old_map = map_dict.get(om_id, {})
                if not old_map:
                    continue
                validated = False
                for chunk_data in chunks:
                    # try to filter out maps that are no longer adjacent to the
                    # rest of the merge set due to other maps not being valid
                    if chunk_data["maps"] and not adjacent_to_set(
                            x, y, chunk_data["maps"]):
                        continue
                    # check that this map's keyed terms match the other keyed
                    # terms in this chunk
                    if validate_old_map(old_map, chunk_data["entry"]):
                        chunk_data["maps"].append({"x": x, "y": y})
                        validated = True
                if not validated:
                    new_entry = copy.deepcopy(basic_entry)
                    chunks.append({"maps": [{"x": x, "y": y}],
                                   "entry": new_entry})

        # then split up any irregular shapes that made it past the screen
        # T and L shapes are possible because every map is adjacent, for
        # instance
        final_chunks = []
        for chunk_data in chunks:
            maps = chunk_data["maps"]
            if len(maps) < 3:
                final_chunks.append(chunk_data)
                continue
            maps.sort(key=itemgetter("x"))
            max_x = maps[-1]["x"]
            min_x = maps[0]["x"]
            maps.sort(key=itemgetter("y"))
            max_y = maps[-1]["y"]
            min_y = maps[0]["y"]
            # if this is a line, square, or rectangle, it's continguous
            if len(maps) == ((max_x - min_x + 1) * (max_y - min_y + 1)):
                final_chunks.append(chunk_data)
                continue
            # if not, just break it into individual maps
            for coords in maps:
                final_chunks.append({
                    "maps": [{"x": coords["x"], "y": coords["y"]}],
                    "entry": copy.deepcopy(chunk_data["entry"])
                })

        if not final_chunks:
            continue

        # debug: do the final chunks look sane?
        #print("chunks: {}".format(json.dumps(chunks, indent=2)))
        # go through the final chunks and merge them
        for chunk_data in final_chunks:
            new_rows = []
            entry = chunk_data["entry"]
            maps = chunk_data["maps"]
            if not maps:
                continue
            first_x = maps[0]["x"]
            first_y = maps[0]["y"]
            for coords in maps:
                x = coords["x"]
                y = coords["y"]
                row_offset = 24 * (y - first_y)
                col_offset = 24 * (x - first_x)
                om_id = mapset.get(x_y_simple(x, y), "")
                old_map = map_dict.get(om_id, {})
                old_obj = old_map.get("object", {})
                old_rows = old_obj.get("rows", {})
                # merge the rows, obviously
                for i in range(0, 24):
                    if not col_offset:
                        new_rows.append(old_rows[i])
                    else:
                        new_rows[i + row_offset] = \
                            "".join([new_rows[i + row_offset], old_rows[i]])
                if len(maps) == 1:
                    entry["om_terrain"] = om_id
                else:
                    if len(entry["om_terrain"]) < (y - first_y + 1):
                        entry["om_terrain"].append([om_id])
                    else:
                        entry["om_terrain"][y - first_y].append(om_id)
                # adjust the offsets for place_ entries before potentially
                # converting set entries
                for place_term in PLACE_TERMS:
                    new_terms = adjust_place(place_term, old_obj, col_offset,
                                             row_offset)
                    if new_terms:
                        entry["object"].setdefault(place_term, [])
                        for term_entry in new_terms:
                            entry["object"][place_term].append(term_entry)
            # finally done with the chunk, so add it to the list
            entry["object"]["rows"] = new_rows
            new_mapgen.append(entry)
            # debug: make sure sure the final chunk is correct
            #print("{}".format(json.dumps(entry, indent=2)))

if output_name:
    with open(output_name, 'w', encoding="utf-8") as output_file:
        output_file.write(json.dumps(new_mapgen))
    exit()

print("{}".format(json.dumps(new_mapgen, indent=2)))



from ..write_text import write_text


def parse_mod_migration(json, origin):
    id = json["id"]

    if "removal_reason" in json:
        write_text(json["removal_reason"], origin,
                   comment="Explaination for removal of mod \"{}\"".format(id))



#!/usr/bin/env python3
"""
Fixes compilation database used by run-clang-tidy.py on Windows.
"""

import getopt
import json
import os
import re
import shlex
import sys


def print_help():
    print("\n"
          "Fix compilation database used by run-clang-tidy.py on Windows.\n"
          "\n"
          "    --help              prints this message\n"
          "    --compile-db=<path> specify the path to compilation database.\n"
          "                        Defaults to build/compile_commands.json\n")


def main(argv):
    compile_db = "build/compile_commands.json"

    try:
        opts, args = getopt.getopt(argv, "", ["help", "compile-db="])
    except getopt.GetoptError:
        print_help()
        return

    for opt, arg in opts:
        if opt == "--help":
            print_help()
            return
        if opt == "--compile-db":
            compile_db = arg

    starts_with_drive_letter = re.compile("^/(\\w)/(.*)$")

    data = None
    with open(compile_db, 'r', encoding="utf-8") as fs:
        data = json.load(fs)

        for j in range(len(data)):
            directory = data[j]["directory"]
            command = shlex.split(data[j]["command"])
            i = 0
            while i < len(command):
                if command[i].startswith("@"):
                    rsp_path = os.path.join(directory, command[i][1:])
                    with open(rsp_path, 'r', encoding="utf-8") as rsp:
                        newflags = shlex.split(rsp.read())
                        command = command[:i] + newflags + command[i + 1:]
                        i = i + len(newflags)
                else:
                    match_result = starts_with_drive_letter.match(command[i])
                    if match_result:
                        command[i] = "{}:/{}".format(match_result.group(1),
                                                     match_result.group(2))
                    i = i + 1
            data[j]["command"] = " ".join([shlex.quote(s) for s in command])

    with open(compile_db, 'w', encoding="utf-8") as fs:
        json.dump(data, fs, indent=2)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3

import json
import argparse
import copy


def hash_key(key):
    if isinstance(key, list):
        return "list_" + "".join(key)
    else:
        return key


def parse_furnter(om_objs, palette, conflicts):
    for map_key, map_val in om_objs.items():
        if isinstance(map_val, list) and len(map_val) == 1:
            map_val = map_val[0]
        pal_val = palette.get(map_key, "")
        if not pal_val:
            palette[map_key] = map_val
        elif pal_val != map_val:
            map_hash = hash_key(map_val)
            pal_hash = hash_key(pal_val)
            conflicts.setdefault(map_key, {})
            conflicts[map_key].setdefault(
                map_hash, {"val": map_val, "count": 0})
            conflicts[map_key].setdefault(
                pal_hash, {"val": pal_val, "count": 0})
            conflicts[map_key][map_hash]["count"] += 1
            conflicts[map_key][pal_hash]["count"] += 1


def decide_conflicts(palette, conflicts):
    for con_key, conflict in conflicts.items():
        most_val = ""
        most_count = -1
        for _, data in conflict.items():
            if data["count"] > most_count:
                most_count = data["count"]
                most_val = data["val"]
        palette[con_key] = most_val


def resolve_conflicts(om_objs, palette, conflicts):
    temp_objs = copy.deepcopy(om_objs)
    for map_key in om_objs:
        map_val = om_objs[map_key]
        if isinstance(map_val, list) and len(map_val) == 1:
            map_val = map_val[0]
        pal_val = palette.get(map_key, "")
        if pal_val == map_val:
            del temp_objs[map_key]
    return temp_objs


args = argparse.ArgumentParser(
    description="Read all the terrain and furniture definitions from a "
                "mapgen .json file and move them into a palette file.  "
                "Symbols with multiple definitions put the most common "
                "definition in the the palette and leave the others in "
                "the mapgen file.")
args.add_argument("mapgen_source", action="store",
                  help="specify json file to convert to palette.")
args.add_argument("palette_name", action="store",
                  help="specify the name of the resulting palette.")
args.add_argument("palette_path", action="store",
                  help="specify the folder for the palette.")
argsDict = vars(args.parse_args())

mapgen = []
mapgen_source = argsDict.get("mapgen_source", "")
palette_name = argsDict.get("palette_name", "")
palette_source = \
    argsDict.get("palette_path", "") + "/" + palette_name + ".json"
if mapgen_source.endswith(".json"):
    try:
        with open(mapgen_source, encoding="utf-8") as mapgen_file:
            mapgen += json.load(mapgen_file)
    except FileNotFoundError:
        exit("Failed: could not find {}".format(mapgen_source))
else:
    exit("Failed: invalid mapgen file name {}".format(mapgen_source))

furn_pal = {}
furn_conflicts = {}
ter_pal = {}
ter_conflicts = {}

# first pass, go through and find all the existing terrains and furnitures and
# any conflicts
for om_tile in mapgen:
    om_object = om_tile.get("object", {})
    om_furn = om_object.get("furniture", {})
    parse_furnter(om_furn, furn_pal, furn_conflicts)
    om_ter = om_object.get("terrain", {})
    parse_furnter(om_ter, ter_pal, ter_conflicts)

# if there are conflicts, pick the most common version for the palette
if furn_conflicts:
    decide_conflicts(furn_pal, furn_conflicts)
if ter_conflicts:
    decide_conflicts(ter_pal, ter_conflicts)

# second pass, remove entries in the palette
for om_tile in mapgen:
    om_object = om_tile.get("object", {})
    om_object.setdefault("palettes", [])
    om_object["palettes"].append(palette_name)
    if om_object.get("furniture"):
        if furn_conflicts:
            om_furn = om_object.get("furniture", {})
            om_furn_final = resolve_conflicts(om_furn, furn_pal,
                                              furn_conflicts)
            if om_furn_final:
                om_object["furniture"] = om_furn_final
            else:
                del om_object["furniture"]
        else:
            del om_object["furniture"]

    if om_object.get("terrain"):
        if ter_conflicts:
            om_ter = om_object.get("terrain", {})
            om_ter_final = resolve_conflicts(om_ter, ter_pal, ter_conflicts)
            if om_ter_final:
                om_object["terrain"] = om_ter_final
            else:
                del om_object["terrain"]
        else:
            del om_object["terrain"]

with open(mapgen_source, 'w', encoding="utf-8") as mapgen_file:
    mapgen_file.write(json.dumps(mapgen, indent=2))

palette_json = [
    {
        "type": "palette",
        "id": palette_name,
        "furniture": furn_pal,
        "terrain": ter_pal
    }
]

with open(palette_source, 'w', encoding="utf-8") as palette_file:
    palette_file.write(json.dumps(palette_json, indent=2))

#print("furniture palette {}".format(json.dumps(furn_pal, indent=2)))
#print("terrain palette {}".format(json.dumps(ter_pal, indent=2)))



#!/usr/bin/env python3

import argparse
import json
import os

args = argparse.ArgumentParser()
args.add_argument("dir", action="store", help="specify json directory")
args_dict = vars(args.parse_args())


def to_resist_vals(jo):
    res = dict()
    if "armor_bash" in jo:
        res["bash"] = jo.pop("armor_bash")
    if "armor_cut" in jo:
        res["cut"] = jo.pop("armor_cut")
    if "armor_stab" in jo:
        res["stab"] = jo.pop("armor_stab")
    if "armor_acid" in jo:
        res["acid"] = jo.pop("armor_acid")
    if "armor_fire" in jo:
        res["heat"] = jo.pop("armor_fire")
    if "armor_bullet" in jo:
        res["bullet"] = jo.pop("armor_bullet")
    if "armor_elec" in jo:
        res["electric"] = jo.pop("armor_elec")
    if "armor_biological" in jo:
        res["biological"] = jo.pop("armor_biological")
    if "armor_cold" in jo:
        res["cold"] = jo.pop("armor_cold")
    if "armor_pure" in jo:
        res["pure"] = jo.pop("armor_pure")

    if res:
        jo["armor"] = res

    if "proportional" in jo:
        jo["proportional"] = to_resist_vals(jo["proportional"])

    if "relative" in jo:
        jo["relative"] = to_resist_vals(jo["relative"])

    return jo


def gen_new(path):
    change = False
    # Having troubles with this script halting?
    # Uncomment below to find the file it's halting on
    #print(path)
    with open(path, "r", encoding="utf-8") as json_file:
        try:
            json_data = json.load(json_file)
        except json.JSONDecodeError:
            print(f"Json Decode Error at {path}")
            print("Ensure that the file is a JSON file consisting of"
                  " an array of objects!")
            return None

        for jo in json_data:
            if type(jo) is not dict:
                print(f"Incorrectly formatted JSON data at {path}")
                print("Ensure that the file is a JSON file consisting"
                      " of an array of objects!")
                break

            if ("type" not in jo or jo["type"] != "MONSTER"):
                continue

            new_data = to_resist_vals(jo)
            if new_data is not None:
                jo = new_data
                change = True

    return json_data if change else None


for root, directories, filenames in os.walk(args_dict["dir"]):
    for filename in filenames:
        if not filename.endswith(".json"):
            continue

        path = os.path.join(root, filename)
        new = gen_new(path)
        if new is not None:
            with open(path, "w", encoding="utf-8") as jf:
                json.dump(new, jf, ensure_ascii=False)
            os.system(f"./tools/format/json_formatter.cgi {path}")



from ..write_text import write_text


def parse_addiction_type(json, origin):
    id = json["id"]
    if "name" in json:
        write_text(json["name"], origin,
                   comment="Name of the \"{}\" addiction's "
                   "effect as it appears in the player's status".format(id))

    if "type_name" in json:
        write_text(json["type_name"], origin,
                   comment="The name of the \"{}\" addiction's source"
                   .format(id))

    if "description" in json:
        write_text(json["description"], origin, c_format=False,
                   comment="Description of the \"{}\" addiction's effects as "
                   "it appears in the player's status".format(id))



from ..helper import get_singular_name
from .effect import parse_effect
from ..write_text import write_text


def parse_enchant(json, origin):
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin,
                   comment="Name of a enchantment")

    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of enchantment \"{}\"".format(name))

    if "hit_me_effect" in json:
        parse_effect(json["hit_me_effect"], origin)

    if "special_vision" in json:
        for vision in json["special_vision"]:
            if "descriptions" in vision:
                for description in vision["descriptions"]:
                    parse_description(description, origin)


def parse_description(description, origin):
    if "text" in description:
        special_vision_id = description["id"]
        write_text(description["text"], origin,
                   comment="Description of creature revealed by special "
                   "vision \"{}\"".format(special_vision_id))



from ..helper import get_singular_name
from ..write_text import write_text


def parse_material(json, origin):
    name = get_singular_name(json["name"])
    write_text(name, origin, comment="Name of material")

    if "bash_dmg_verb" in json:
        write_text(json["bash_dmg_verb"], origin,
                   comment="Bash damage verb of material")

    if "cut_dmg_verb" in json:
        write_text(json["cut_dmg_verb"], origin,
                   comment="Cut damage verb of material")

    if "dmg_adj" in json:
        for i in range(4):
            write_text(json["dmg_adj"][i], origin,
                       comment="Damage adjective of material")



from ..helper import get_singular_name
from .use_action import parse_use_action
from ..write_text import write_text


def parse_gun(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Name of a gun", plural=True)
    elif "id" in json:
        name = json["id"]

    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of gun \"{}\"".format(name))

    if "use_action" in json:
        parse_use_action(json["use_action"], origin, name)

    if "variants" in json:
        for variant in json["variants"]:
            variant_name = get_singular_name(variant["name"])
            write_text(variant["name"], origin,
                       comment="Variant name of gun \"{}\"".format(name),
                       plural=True)
            write_text(variant["description"], origin,
                       comment="Description of variant \"{0}\" of gun \"{1}\""
                       .format(name, variant_name))

    if "modes" in json:
        for mode in json["modes"]:
            write_text(mode[1], origin,
                       comment="Firing mode of gun")

    if "skill" in json:
        if json["skill"] != "archery":
            write_text(json["skill"], origin, context="gun_type_type",
                       comment="Skill associated with gun")

    if "reload_noise" in json:
        write_text(json["reload_noise"], origin,
                   comment="Reload noise of gun \"{}\"".format(name))

    if "valid_mod_locations" in json:
        for loc in json["valid_mod_locations"]:
            write_text(loc[0], origin,
                       comment="Valid mod location of gun")



#!/usr/bin/env python3

import re
import urllib.request
import urllib.parse
import urllib.error
import json
import itertools
import logging
import threading
import collections
import time
import os
import sys
import argparse
import pathlib
import contextlib
import xml.etree.ElementTree
from datetime import time as dtime, date, datetime, timedelta


log = logging.getLogger('generate_changelog')


class MissingCommitException(Exception):
    pass


class JenkinsBuild:
    """Representation of a Jenkins Build"""

    def __init__(self, number, last_hash, branch, build_dttm, is_building,
                 build_result, block_ms, wait_ms, build_ms):
        self.number = number
        self.last_hash = last_hash
        self.branch = branch
        self.build_dttm = build_dttm
        self.is_building = is_building
        self.build_result = build_result
        self.wait_ms = wait_ms
        self.build_ms = build_ms
        self.block_ms = block_ms

    def was_successful(self):
        return not self.is_building and self.build_result == 'SUCCESS'

    def __str__(self):
        return (f'{self.__class__.__name__}[{self.number} - '
                f'{self.last_hash} - {self.build_dttm} - '
                f'{self.build_result}]')


class Commit(object):
    """Representation of a generic GitHub Commit"""

    def __init__(self, hash_id, message, commit_dttm, author, parents):
        self.hash = hash_id
        self.message = message
        self.commit_dttm = commit_dttm
        self.author = author
        self.parents = parents

    @property
    def commit_date(self):
        return self.commit_dttm.date()

    def committed_after(self, commit_dttm):
        return self.commit_dttm is not None and self.commit_dttm >= commit_dttm

    def committed_before(self, commit_dttm):
        return self.commit_dttm is not None and self.commit_dttm <= commit_dttm

    @property
    def is_merge(self):
        return len(self.parents) > 1

    def __str__(self):
        return (f'{self.__class__.__name__}[{self.hash} - '
                f'{self.commit_dttm} - {self.message} BY {self.author}]')


class PullRequest(object):
    """Representation of a generic GitHub Pull Request"""

    def __init__(self, pr_id, title, author, state, body, merge_hash,
                 merge_dttm, update_dttm):
        self.id = pr_id
        self.author = author
        self.title = title
        self.body = body
        self.state = state
        self.merge_hash = merge_hash
        self.merge_dttm = merge_dttm
        self.update_dttm = update_dttm

    @property
    def update_date(self):
        return self.update_dttm.date()

    @property
    def merge_date(self):
        return self.merge_dttm.date()

    @property
    def is_merged(self):
        return self.merge_dttm is not None

    def merged_after(self, merge_dttm):
        return self.merge_dttm is not None and self.merge_dttm >= merge_dttm

    def merged_before(self, merge_dttm):
        return self.merge_dttm is not None and self.merge_dttm <= merge_dttm

    def updated_after(self, update_dttm):
        return self.update_dttm is not None and self.update_dttm >= update_dttm

    def updated_before(self, update_dttm):
        return self.update_dttm is not None and self.update_dttm <= update_dttm

    def __str__(self):
        return (f'{self.__class__.__name__}[{self.id} - {self.merge_dttm} - '
                f'{self.title} BY {self.author}]')


class SummaryType:
    """Different valid Summary Types. Intended to be used as a enum/constant
    class, no instantiation needed."""

    NONE = 'NONE'
    FEATURES = 'FEATURES'
    CONTENT = 'CONTENT'
    INTERFACE = 'INTERFACE'
    MODS = 'MODS'
    BALANCE = 'BALANCE'
    BUGFIXES = 'BUGFIXES'
    PERFORMANCE = 'PERFORMANCE'
    INFRASTRUCTURE = 'INFRASTRUCTURE'
    BUILD = 'BUILD'
    I18N = 'I18N'


class CDDAPullRequest(PullRequest):
    """A Pull Request with logic specific to CDDA Repository and their
    "Summary" descriptions"""

    SUMMARY_REGEX = re.compile(
        r'(?i:####\sSummary)\s*'
        r'`*(?i:SUMMARY:?\s*)?(?P<pr_type>\w+)\s*(?:"(?P<pr_desc>.+)")?',
        re.MULTILINE)

    VALID_SUMMARY_CATEGORIES = (
        'Content',
        'Features',
        'Interface',
        'Mods',
        'Balance',
        'I18N',
        'Bugfixes',
        'Performance',
        'Build',
        'Infrastructure',
        'None',
    )

    EXAMPLE_SUMMARIES_IN_TEMPLATE = (
        ("Category", "description"),
        ("Category", "Brief description"),
        ("Content", "Adds new mutation category 'Mouse'"),
    )

    def __init__(self, pr_id, title, author, state, body, merge_hash,
                 merge_dttm, update_dttm, store_body=False):
        super().__init__(
            pr_id, title, author, state, body if store_body else '',
            merge_hash, merge_dttm, update_dttm)
        self.summ_type, self.summ_desc = self._get_summary(body)

    @property
    def summ_type(self):
        return self._summ_type

    @summ_type.setter
    def summ_type(self, value):
        self._summ_type = value.upper() if value is not None else None

    @property
    def has_valid_summary(self):
        return (self.summ_type == SummaryType.NONE or
                (self.summ_type is not None and self.summ_desc is not None))

    def _get_summary(self, body):
        matches = list(re.finditer(self.SUMMARY_REGEX, body))
        # Fix weird cases where a PR have multiple SUMMARY
        # coming mostly from template lines that weren't removed by the pull
        # requester.  For example:
        # https://api.github.com/repos/CleverRaven/Cataclysm-DDA/pulls/25604

        def summary_filter(x):
            return ((x.group('pr_type'), x.group('pr_desc')) not in
                    self.EXAMPLE_SUMMARIES_IN_TEMPLATE)
        matches = list(filter(summary_filter, matches))
        if len(matches) > 1:
            log.warning(f'More than one SUMMARY defined in PR {self.id}!')

        match = matches[0] if matches else None
        upper_cats = (x.upper() for x in self.VALID_SUMMARY_CATEGORIES)
        if match is None or match.group('pr_type').upper() not in upper_cats:
            return None, None
        else:
            return match.group('pr_type'), match.group('pr_desc')

    def __str__(self):
        if self.has_valid_summary and self.summ_type == SummaryType.NONE:
            return (f'{self.__class__.__name__}'
                    f'[{self.id} - {self.merge_dttm} - {self.summ_type} - '
                    f'{self.title} BY {self.author}]')
        elif self.has_valid_summary:
            return (f'{self.__class__.__name__}'
                    f'[{self.id} - {self.merge_dttm} - {self.summ_type} - '
                    f'{self.summ_desc} BY {self.author}]')
        else:
            return (f'{self.__class__.__name__}'
                    f'[{self.id} - {self.merge_dttm} - '
                    f'{self.title} BY {self.author}]')


class JenkinsBuildFactory:
    """Abstraction for instantiation of new Commit objects"""

    def create(self, number, last_hash, branch, build_dttm, is_building,
               build_result, block_ms, wait_ms, build_ms):
        return JenkinsBuild(number, last_hash, branch, build_dttm, is_building,
                            build_result, block_ms, wait_ms, build_ms)


class CommitFactory:
    """Abstraction for instantiation of new Commit objects"""

    def create(self, hash_id, message, commit_date, author, parents):
        return Commit(hash_id, message, commit_date, author, parents)


class CDDAPullRequestFactory:
    """Abstraction for instantiation of new CDDAPullRequests objects"""

    def __init__(self, store_body=False):
        self.store_body = store_body

    def create(self, pr_id, title, author, state, body, merge_hash, merge_dttm,
               update_dttm):
        return CDDAPullRequest(pr_id, title, author, state, body, merge_hash,
                               merge_dttm, update_dttm, self.store_body)


class CommitRepository:
    """Groups Commits for storage and common operations"""

    def __init__(self):
        self.ref_by_commit_hash = {}
        self._latest_commit = None

    def add(self, commit):
        is_newer = (self._latest_commit is None or
                    self._latest_commit.commit_date < commit.commit_date)
        if is_newer:
            self._latest_commit = commit
        self.ref_by_commit_hash[commit.hash] = commit

    def add_multiple(self, commits):
        for commit in commits:
            self.add(commit)

    def get_commit(self, commit_hash):
        if commit_hash in self.ref_by_commit_hash:
            return self.ref_by_commit_hash[commit_hash]
        return None

    def get_latest_commit(self):
        return self._latest_commit

    def get_all_commits(self, filter_by=None, sort_by=None):
        """Return all Commits in Repository. No order is guaranteed."""
        commit_list = self.ref_by_commit_hash.values()

        if filter_by is not None:
            commit_list = filter(filter_by, commit_list)

        if sort_by is not None:
            commit_list = sorted(commit_list, key=sort_by)

        for commit in commit_list:
            yield commit

    def traverse_commits_by_first_parent(self, initial_hash=None):
        """Iterate through Commits connected by the first Parent, until a
        parent is not found in the Repository

        This is like using 'git log --first-parent $commit', which avoids
        Commits "inside" Pull Requests / Merges.
        But returns Merge Commits (which often can be related to a single PR)
        and Commits directly added to a Branch.
        """
        if initial_hash is not None:
            commit = self.get_commit(initial_hash)
        else:
            commit = self.get_latest_commit()

        while commit is not None:
            yield commit
            commit = self.get_commit(commit.parents[0])

    def get_commit_range_by_hash(self, latest_hash, oldest_hash):
        """Return Commits between initial_hash (including) and oldest_hash
        (excluding) connected by the first Parent."""
        for commit in self.traverse_commits_by_first_parent(latest_hash):
            if commit.hash == oldest_hash:
                return
            yield commit
        # consumed the whole commit list in the Repo and didn't find
        # oldest_hash so returned commit list is incomplete
        raise MissingCommitException(
            "Can't generate commit list for specified hash range."
            " There are missing Commits in CommitRepository."
        )

    def get_commit_range_by_date(self, latest_dttm, oldest_dttm):
        """Return Commits between latest_dttm (including) and oldest_dttm
        (excluding) connected by the first Parent."""
        for commit in self.traverse_commits_by_first_parent():
            # assuming that traversing by first parent have chronological order
            # (should be DESC BY commit_dttm)
            if commit.commit_dttm <= oldest_dttm:
                return
            if latest_dttm >= commit.commit_dttm > oldest_dttm:
                yield commit
        # consumed the whole commit list and didn't find a commit past our
        # build.  So returned commit list *could be* incomplete
        raise MissingCommitException(
            "Can't generate commit list for specified date range."
            " There are missing Commits in CommitRepository."
        )

    def purge_references(self):
        self.ref_by_commit_hash.clear()


class CDDAPullRequestRepository:
    """Groups Pull Requests for storage and common operations"""

    def __init__(self):
        self.ref_by_pr_id = {}
        self.ref_by_merge_hash = {}

    def add(self, pull_request):
        self.ref_by_pr_id[pull_request.id] = pull_request
        if pull_request.merge_hash is not None:
            self.ref_by_merge_hash[pull_request.merge_hash] = pull_request

    def add_multiple(self, pull_requests):
        for pr in pull_requests:
            self.add(pr)

    def get_pr_by_merge_hash(self, merge_hash):
        if merge_hash in self.ref_by_merge_hash:
            return self.ref_by_merge_hash[merge_hash]
        return None

    def get_all_pr(self, filter_by=None, sort_by=None):
        """Return all Pull Requests in Repository. No order is guaranteed."""
        pr_list = self.ref_by_pr_id.values()

        if filter_by is not None:
            pr_list = filter(filter_by, pr_list)

        if sort_by is not None:
            pr_list = sorted(pr_list, key=sort_by)

        for pr in pr_list:
            yield pr

    def get_merged_pr_list_by_date(self, latest_dttm, oldest_dttm):
        """Return PullRequests merged between latest_dttm (including) and
        oldest_dttm (excluding)."""
        for pr in self.get_all_pr(filter_by=lambda x: x.is_merged,
                                  sort_by=lambda x: -x.merge_dttm.timestamp()):
            if latest_dttm >= pr.merge_dttm > oldest_dttm:
                yield pr

    def purge_references(self):
        self.ref_by_merge_hash.clear()


class JenkinsBuildRepository:
    """Groups JenkinsBuilds for storage and common operations"""

    def __init__(self):
        self.ref_by_build_number = {}

    def add(self, build):
        if build.number is not None:
            self.ref_by_build_number[build.number] = build

    def add_multiple(self, builds):
        for build in builds:
            self.add(build)

    def get_build_by_number(self, build_number):
        if build_number in self.ref_by_build_number:
            return self.ref_by_build_number[build_number]
        return None

    def get_previous_build(self, build_number, condition=lambda x: True):
        for x in range(build_number - 1, 0, -1):
            prev_build = self.get_build_by_number(x)
            if prev_build is not None and condition(prev_build):
                return prev_build
        return None

    def get_previous_successful_build(self, build_number):
        return self.get_previous_build(build_number,
                                       lambda x: x.was_successful())

    def get_all_builds(self, filter_by=None, sort_by=None):
        """Return all Builds in Repository. No order is guaranteed."""
        build_list = self.ref_by_build_number.values()

        if filter_by is not None:
            build_list = filter(filter_by, build_list)

        if sort_by is not None:
            build_list = sorted(build_list, key=sort_by)

        for build in build_list:
            yield build

    def purge_references(self):
        self.ref_by_build_number.clear()


class JenkinsApi:

    JENKINS_BUILD_LIST_API = \
        r'http://gorgon.narc.ro:8080/job/Cataclysm-Matrix/api/xml'

    JENKINS_BUILD_LONG_LIST_PARAMS = {
        'tree': r'allBuilds[number,timestamp,building,result,actions[buildingDurationMillis,waitingDurationMillis,blockedDurationMillis,lastBuiltRevision[branch[name,SHA1]]]]',  # noqa: E501
        'xpath': r'//allBuild',
        'wrapper': r'allBuilds',
        'exclude': r'//action[not(buildingDurationMillis|waitingDurationMillis|blockedDurationMillis|lastBuiltRevision/branch)]'  # noqa: E501
    }

    JENKINS_BUILD_SHORT_LIST_PARAMS = {
        'tree': r'builds[number,timestamp,building,result,actions[buildingDurationMillis,waitingDurationMillis,blockedDurationMillis,lastBuiltRevision[branch[name,SHA1]]]]',  # noqa: E501
        'xpath': r'//build',
        'wrapper': r'builds',
        'exclude': r'//action[not(buildingDurationMillis|waitingDurationMillis|blockedDurationMillis|lastBuiltRevision/branch)]'  # noqa: E501
    }

    def __init__(self, build_factory):
        self.build_factory = build_factory

    def get_build_list(self):
        """Return the builds from Jenkins. API limits the result to last 999
        builds."""
        request_url = (self.JENKINS_BUILD_LIST_API + '?' +
                       urllib.parse.urlencode(
                           self.JENKINS_BUILD_LONG_LIST_PARAMS))
        api_request = urllib.request.Request(request_url)
        log.debug(f'Processing Request {api_request.full_url}')
        with urllib.request.urlopen(api_request) as api_response:
            api_data = xml.etree.ElementTree.fromstring(api_response.read())
        log.debug('Jenkins Request DONE!')

        for build_data in api_data:
            yield self._create_build_from_api_data(build_data)

    def _create_build_from_api_data(self, build_data):
        """Create a JenkinsBuild instance based on data from Jenkins API"""
        jb_number = int(build_data.find('number').text)
        jb_build_dttm = datetime.utcfromtimestamp(
            int(build_data.find('timestamp').text) // 1000)
        jb_is_building = build_data.find('building').text == 'true'

        jb_block_ms = timedelta(0)
        jb_wait_ms = timedelta(0)
        jb_build_ms = timedelta(0)
        jb_build_result = None
        if not jb_is_building:
            jb_build_result = build_data.find('result').text
            jb_block_ms = timedelta(milliseconds=int(
                build_data.find(r'.//action/blockedDurationMillis').text))
            jb_wait_ms = timedelta(milliseconds=int(
                build_data.find(r'.//action/waitingDurationMillis').text))
            jb_build_ms = timedelta(milliseconds=int(
                build_data.find(r'.//action/buildingDurationMillis').text))

        jb_last_hash = None
        jb_branch = None
        sha1 = build_data.find(r'.//action/lastBuiltRevision/branch/SHA1')
        if sha1 is not None:
            jb_last_hash = sha1.text
            jb_branch = build_data.find(
                r'.//action/lastBuiltRevision/branch/name').text

        return self.build_factory.create(
            jb_number, jb_last_hash, jb_branch, jb_build_dttm, jb_is_building,
            jb_build_result, jb_block_ms, jb_wait_ms, jb_build_ms)


class CommitApi:

    def __init__(self, commit_factory, api_token):
        self.commit_factory = commit_factory
        self.api_token = api_token

    def get_commit_list(self, min_commit_dttm, max_commit_dttm,
                        branch='master', max_threads=15):
        """Return a list of Commits from specified commit date up to now. Order
        is not guaranteed by threads.

            params:
                min_commit_dttm = None or minimum commit date to be part of the
                    result set (UTC+0 timezone)
                max_commit_dttm = None or maximum commit date to be part of the
                    result set (UTC+0 timezone)
        """
        if min_commit_dttm is None:
            min_commit_dttm = datetime.min
        if max_commit_dttm is None:
            max_commit_dttm = datetime.utcnow()

        results_queue = collections.deque()
        request_generator = CommitApiGenerator(self.api_token, min_commit_dttm,
                                               max_commit_dttm, branch)
        threaded = MultiThreadedGitHubApi()
        threaded.process_api_requests(
            request_generator,
            self._process_commit_data_callback(results_queue),
            max_threads=max_threads)

        return (commit for commit in results_queue
                if commit.committed_after(min_commit_dttm) and
                commit.committed_before(max_commit_dttm))

    def _process_commit_data_callback(self, results_queue):
        """Returns a callback that will process data into Commits instances and
        stop threads when needed."""
        def _process_commit_data_callback_closure(json_data, request_generator,
                                                  api_request):
            nonlocal results_queue
            commit_list = [
                self._create_commit_from_api_data(x) for x in json_data]
            for commit in commit_list:
                results_queue.append(commit)

            if request_generator.is_active and len(commit_list) == 0:
                log.debug('Target page found, stop giving threads more '
                          f'requests. Target URL: {api_request.full_url}')
                request_generator.deactivate()
        return _process_commit_data_callback_closure

    def _create_commit_from_api_data(self, commit_data):
        """Create a Commit instance based on data from GitHub API"""
        if commit_data is None:
            return None
        commit_sha = commit_data['sha']
        # some commits have null in author.login :S like:
        # https://api.github.com/repos/CleverRaven/Cataclysm-DDA/commits/569bef1891a843ec71654530a64d51939aabb3e2
        # I try to use author.login when possible or fallback to
        # "commit.author" which doesn't match with usernames in pull requests
        # API (I guess it comes from the distinction between "name" and
        # "username".
        # I rather have a name that doesn't match that leave it empty.
        # Anyways, I'm surprised but GitHub API sucks, is super inconsistent
        # and not well thought or documented.
        if commit_data['author'] is not None:
            if 'login' in commit_data['author']:
                commit_author = commit_data['author']['login']
            else:
                commit_author = 'null'
        else:
            commit_author = commit_data['commit']['author']['name']
        if commit_data['commit']['message']:
            commit_message = commit_data['commit']['message'].splitlines()[0]
        else:
            commit_message = ''
        commit_dttm = commit_data['commit']['committer']['date']
        if commit_dttm:
            commit_dttm = datetime.fromisoformat(commit_dttm.rstrip('Z'))
        else:
            commit_dttm = None
        commit_parents = tuple(p['sha'] for p in commit_data['parents'])

        return self.commit_factory.create(
            commit_sha, commit_message, commit_dttm, commit_author,
            commit_parents)


class PullRequestApi:

    GITHUB_API_SEARCH = r'https://api.github.com/search/issues'

    def __init__(self, pr_factory, api_token):
        self.pr_factory = pr_factory
        self.api_token = api_token

    def search_for_pull_request(self, commit_hash):
        """Returns the Pull Request ID of a specific Commit Hash using a Search
        API in GitHub

        AVOID AT ALL COST: You have to check Commit by Commit and is super
        slow!

        Based on
        https://help.github.com/articles/searching-issues-and-pull-requests/#search-by-commit-sha
        No need to make this multi-threaded as GitHub limits this to 30
        requests per minute.
        """
        params = {
            'q': 'is:pr is:merged repo:CleverRaven/Cataclysm-DDA '
                 f'{commit_hash}'
        }
        request_builder = GitHubApiRequestBuilder(self.api_token)
        api_request = request_builder.create_request(self.GITHUB_API_SEARCH,
                                                     params)

        api_data = do_github_request(api_request)

        # I found some cases where the search found multiple PRs because of a
        # reference of the commit hash in another PR description. But it had a
        # huge difference in score, it seems the engine scores over 100 the one
        # that actually has the commit hash as "Merge Commit SHA".
        # Output example:
        # https://api.github.com/search/issues
        # ?q=is%3Apr+is%3Amerged+repo%3ACleverRaven%2FCataclysm-DDA+f0c6908d154cd0fb190c2116de2bf2d3131458c3

        # If the API don't return a score of 100 or more for the highest score
        # item (the first item), then the commit is probably not part of any
        # pull request.  This assumption was backed up by checking ~9 months of
        # commits
        if api_data['total_count'] == 0 or api_data['items'][0]['score'] < 100:
            return None
        else:
            return api_data['items'][0]['number']

    def get_pr_list(self, min_dttm, max_dttm=None, state='all',
                    merged_only=False, max_threads=15):
        """Return a list of PullRequests from specified commit date up to now.
        Order is not guaranteed by threads.

            params:
                min_dttm = None or minimum update date on the PR to be part of
                    the result set (UTC+0 timezone)
                max_dttm = None or maximum update date on the PR to be part of
                    the result set (UTC+0 timezone)
                state = 'open' | 'closed' | 'all'
                merged_only = search only 'closed' state PRs, and filter PRs by
                    merge date instead of update date
        """
        if min_dttm is None:
            min_dttm = datetime.min
        if max_dttm is None:
            max_dttm = datetime.utcnow()
        if merged_only:
            state = 'closed'

        results_queue = collections.deque()
        request_generator = PullRequestApiGenerator(self.api_token, state)
        threaded = MultiThreadedGitHubApi()
        threaded.process_api_requests(
            request_generator,
            self._process_pr_data_callback(results_queue, min_dttm),
            max_threads=max_threads)

        if merged_only:
            return (pr for pr in results_queue
                    if pr.is_merged and pr.merged_after(min_dttm) and
                    pr.merged_before(max_dttm))
        else:
            return (pr for pr in results_queue
                    if pr.updated_after(min_dttm) and
                    pr.updated_before(max_dttm))

    def _process_pr_data_callback(self, results_queue, min_dttm):
        """Returns a callback that will process data into Pull Requests objects
        and stop threads when needed."""
        def _process_pr_data_callback_closure(json_data, request_generator,
                                              api_request):
            nonlocal results_queue, min_dttm
            pull_request_list = [
                self._create_pr_from_api_data(x) for x in json_data]
            for pr in pull_request_list:
                results_queue.append(pr)

            # this check on update date makes sure we get the GitHub API pages
            # we need a more precise PR filter of the result is made from
            # results_queue later
            target_page_found = not any(
                pr.updated_after(min_dttm) for pr in pull_request_list)

            if len(pull_request_list) == 0 or target_page_found:
                if request_generator.is_active:
                    log.debug(
                        'Target page found, stop giving threads more '
                        f'requests. Target URL: {api_request.full_url}')
                    request_generator.deactivate()

        return _process_pr_data_callback_closure

    def _create_pr_from_api_data(self, pr_data):
        """Create a PullRequest instance based on data from GitHub API"""
        if pr_data is None:
            return None
        pr_number = pr_data['number']
        pr_author = pr_data['user']['login']
        pr_title = pr_data['title']
        pr_state = pr_data['state']
        pr_merge_hash = pr_data['merge_commit_sha']
        # python expects an ISO date with the Z to properly parse it, so I
        # strip it.
        if pr_data['merged_at']:
            pr_merge_dttm = datetime.fromisoformat(
                pr_data['merged_at'].rstrip('Z'))
        else:
            pr_merge_dttm = None
        if pr_data['updated_at']:
            pr_update_dttm = datetime.fromisoformat(
                pr_data['updated_at'].rstrip('Z'))
        else:
            pr_update_dttm = None
        # PR description can be empty :S example:
        # https://github.com/CleverRaven/Cataclysm-DDA/pull/24213
        pr_body = pr_data['body'] if pr_data['body'] else ''

        return self.pr_factory.create(
            pr_number, pr_title, pr_author, pr_state, pr_body, pr_merge_hash,
            pr_merge_dttm, pr_update_dttm)


class MultiThreadedGitHubApi:

    def process_api_requests(self, request_generator, callback,
                             max_threads=15):
        """Process HTTP API requests on threads and call the callback for each
        result JSON

            params:
                callback = executed when data is available, should expect two
                    params: (json data, request_generator)
        """
        # start threads
        threads = []
        for x in range(1, max_threads + 1):
            t = threading.Thread(
                target=exit_on_exception(
                    self._process_api_requests_on_threads),
                args=(request_generator, callback))
            t.name = f'WORKER_{x:03}'
            threads.append(t)
            t.daemon = True
            t.start()
            time.sleep(0.1)

        # block waiting until threads get all results
        for t in threads:
            t.join()
        log.debug('Threads have finished processing all the required GitHub '
                  'API Requests!!!')

    @staticmethod
    def _process_api_requests_on_threads(request_generator, callback):
        """Process HTTP API requests and call the callback for each result
        JSON"""
        log.debug('Thread Started!')
        api_request = request_generator.generate()
        while api_request is not None:
            callback(do_github_request(api_request), request_generator,
                     api_request)
            api_request = request_generator.generate()
        log.debug('No more requests left, killing Thread.')


class GitHubApiRequestBuilder(object):

    def __init__(self, api_token, timezone='Etc/UTC'):
        self.api_token = api_token
        self.timezone = timezone

    def create_request(self, url, params=None):
        """Creates an API request based on provided URL and GET Parameters"""
        if params is None:
            request_url = url
        else:
            request_url = url + '?' + urllib.parse.urlencode(params)

        if self.api_token is None:
            api_request = urllib.request.Request(request_url)
        else:
            api_headers = {
                'Authorization': 'token ' + self.api_token,
                'Time-Zone': self.timezone,
            }
            api_request = urllib.request.Request(
                request_url, headers=api_headers)

        return api_request


class CommitApiGenerator(GitHubApiRequestBuilder):
    """Generates multiple HTTP requests to get Commits, used from Threads to
    get data until a condition is met."""

    GITHUB_API_LIST_COMMITS = \
        r'https://api.github.com/repos/CleverRaven/Cataclysm-DDA/commits'

    def __init__(self, api_token, since_dttm=None, until_dttm=None,
                 sha='master', initial_page=1, step=1, timezone='Etc/UTC'):
        super().__init__(api_token, timezone)
        self.sha = sha
        self.since_dttm = since_dttm
        self.until_dttm = until_dttm
        self.page = initial_page
        self.step = step
        self.lock = threading.RLock()
        self.is_active = True

    @property
    def is_active(self):
        with self.lock:
            return self._is_active

    @is_active.setter
    def is_active(self, value):
        with self.lock:
            self._is_active = value

    def deactivate(self):
        """Stop generate() from creating new HTTP requests on future calls."""
        self.is_active = False

    def generate(self):
        """Returns an HTTP request to get Commits for a different API result
        page each call until deactivate()."""
        with self.lock:
            if self.is_active:
                req = self.create_request(self.since_dttm, self.until_dttm,
                                          self.sha, self.page)
                self.page += self.step
                return req
            else:
                return None

    def create_request(self, since_dttm=None, until_dttm=None, sha='master',
                       page=1):
        """Creates an HTTP Request to GitHub API to get Commits from CDDA
        repository."""
        params = {
            'sha': sha,
            'page': page,
        }
        if since_dttm is not None:
            params['since'] = since_dttm.isoformat()
        if until_dttm is not None:
            params['until'] = until_dttm.isoformat()

        return super().create_request(self.GITHUB_API_LIST_COMMITS, params)


class PullRequestApiGenerator(GitHubApiRequestBuilder):
    """Generates multiple HTTP requests to get Pull Requests, used from Threads
    to get data until a condition is met."""

    GITHUB_API_LIST_PR = \
        r'https://api.github.com/repos/CleverRaven/Cataclysm-DDA/pulls'

    def __init__(self, api_token, state='all', initial_page=1, step=1,
                 timezone='Etc/UTC'):
        super().__init__(api_token, timezone)
        self.page = initial_page
        self.step = step
        self.state = state
        self.lock = threading.RLock()
        self.is_active = True

    @property
    def is_active(self):
        with self.lock:
            return self._is_active

    @is_active.setter
    def is_active(self, value):
        with self.lock:
            self._is_active = value

    def deactivate(self):
        """Stop generate() from creating new HTTP requests on future calls."""
        self.is_active = False

    def generate(self):
        """Returns an HTTP request to get Pull Requests for a different API
        result page each call until deactivate()."""
        with self.lock:
            if self.is_active:
                req = self.create_request(self.state, self.page)
                self.page += self.step
                return req
            else:
                return None

    def create_request(self, state='all', page=1):
        """Creates an HTTP Request to GitHub API to get Pull Requests from CDDA
        repository.

            params:
                state = 'open' | 'closed' | 'all'
        """
        params = {
            'state': state,
            'sort': 'updated',
            'direction': 'desc',
            'page': page,
        }
        return super().create_request(self.GITHUB_API_LIST_PR, params)


def exit_on_exception(func):
    """Decorator to terminate the main script and all threads if a thread
    generates an Exception"""
    def exit_on_exception_closure(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as err:
            log.exception(f'Unhandled Exception: {err}')
            os._exit(-10)
    return exit_on_exception_closure


def do_github_request(api_request, retry_on_limit=3):
    """Do an HTTP request to GitHub and retries in case of hitting API
    limits"""
    for retry in range(1, retry_on_limit + 2):
        try:
            log.debug(f'Processing Request {api_request.full_url}')
            with urllib.request.urlopen(api_request) as api_response:
                return json.load(api_response)
        except urllib.error.HTTPError as err:
            # hit rate limit, wait and retry
            is_403 = err.code == 403
            is_502 = err.code == 502

            if is_403 and err.getheader('Retry-After'):
                wait = int(err.getheader('Retry-After')) + 5
                log.info(f'Reached GitHub API rate limit. Retry {retry}, '
                         f'waiting {wait} secs...')
                time.sleep(wait)
            elif is_403 and err.getheader('X-RateLimit-Remaining') == '0':
                reset = int(err.getheader('X-RateLimit-Reset'))
                delta = datetime.utcfromtimestamp(reset) - datetime.utcnow()
                wait = delta.seconds + 5
                log.info(f'Reached GitHub API rate limit. Retry {retry}, '
                         f'waiting {wait} secs...')
                time.sleep(wait)
            elif is_502:
                wait = 5
                log.info(f'GitHub API 502 Gateway Error. Retry {retry}, '
                         f'waiting {wait} secs...')
                time.sleep(wait)
            else:
                # other kind of http error, just implode
                log.exception(f'Unhandled Exception: {err} - '
                              f'HTTP Headers: {err.getheaders()}')
                raise
    raise Exception('Retry limit reached')


def read_personal_token(filename):
    """Return Personal Token from specified file, None if no file is provided
    or file doesn't exist.

    Personal Tokens can be generated in https://github.com/settings/tokens
    This makes GitHub API have higher usage limits
    """
    if filename is None:
        return None

    try:
        with open(
                pathlib.Path(str(filename)).expanduser(),
                encoding="utf-8") as token_file:
            match = re.search('(?P<token>\\S+)', token_file.read(),
                              flags=re.MULTILINE)
            if match is not None:
                return match.group('token')
    except IOError:
        pass

    return None


@contextlib.contextmanager
def smart_open(filename=None, *args, **kwargs):
    if filename and (filename == '-' or filename == sys.stdout):
        fh = sys.stdout
    else:
        fh = open(filename, *args, **kwargs)

    try:
        yield fh
    finally:
        if fh is not sys.stdout:
            fh.close()


def validate_file_for_writing(filepath):
    if (filepath is not None and
        filepath != sys.stdout and
            (not filepath.parent.exists() or
             not filepath.parent.is_dir())):
        return False


def main_entry(argv):
    parser = argparse.ArgumentParser(
        description='Generates Changelog from now until the specified data.\n'
                    'generate_changelog.py -D changelog_2019_03 '
                    '-t ../repo_token -f -e 2019-04-01 2019-03-01''',
                    formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument(
        'target_date',
        help='Specify when should stop generating. Accepts "YYYY-MM-DD" ISO '
             '8601 Date format.',
        type=lambda d: datetime.combine(date.fromisoformat(d), dtime.min)
    )

    def convert_path(x):
        if x == '-':
            return sys.stdout
        else:
            return pathlib.Path(x).expanduser().resolve()

    parser.add_argument(
        '-D', '--by-date',
        help='Indicates changes should be presented grouped by Date. Requires '
             'a filename parameter, "-" for STDOUT',
        type=convert_path,
        default=None
    )

    parser.add_argument(
        '-B', '--by-build',
        help='Indicates changes should be presented grouped by Build. '
             'Requires a filename parameter, "-" for STDOUT',
        type=convert_path,
        default=None
    )

    parser.add_argument(
        '-e', '--end-date',
        help='Specify when should start generating. Accepts "YYYY-MM-DD" ISO '
             '8601 Date format.',
        type=lambda d: datetime.combine(date.fromisoformat(d), dtime.max),
        default=datetime.utcnow()
    )

    parser.add_argument(
        '-t', '--token-file',
        help='Specify where to read the Personal Token. Default '
             '"~/.generate_changelog.token".',
        type=lambda x: pathlib.Path(x).expanduser().resolve(),
        default='~/.generate_changelog.token'
    )

    parser.add_argument(
        '-T', '--token-string',
        help='Specify the Personal Token as an inline argument.',
        default=None
    )

    parser.add_argument(
        '-N', '--include-summary-none',
        action='store_true',
        help='Indicates if Pull Requests with Summary "None" should be '
             'included in the output.',
        default=None
    )

    parser.add_argument(
        '-f', '--flatten-output',
        action='store_true',
        help='Output a flattened format with no headings.',
        default=None
    )

    parser.add_argument(
        '--verbose',
        action='store_true',
        help='Indicates the logging system to generate more information about '
             'actions.',
        default=None
    )

    arguments = parser.parse_args(argv[1:])

    logging.basicConfig(
        level=logging.DEBUG if arguments.verbose else logging.INFO,
        format='   LOG | %(threadName)s | %(levelname)s | %(message)s')

    log.debug(f'Commandline Arguments (+defaults): {arguments}')

    if validate_file_for_writing(arguments.by_date):
        raise ValueError("Specified directory in --by-date doesn't exist: "
                         f"{arguments.by_date.parent}")

    if validate_file_for_writing(arguments.by_build):
        raise ValueError("Specified directory in --by-build doesn't exist: "
                         f"{arguments.by_build.parent}")

    personal_token = read_personal_token(arguments.token_file)
    if personal_token is None:
        personal_token = arguments.token_string
        if personal_token is None:
            log.warning("GitHub Token was not provided, API calls will have "
                        "severely limited rates.")

    if arguments.by_date is None and arguments.by_build is None:
        raise ValueError("Script should be called with either --by-date or "
                         "--by-build arguments or both.")

    main_output(arguments.by_date, arguments.by_build, arguments.target_date,
                arguments.end_date, personal_token,
                arguments.include_summary_none, arguments.flatten_output)


def get_github_api_data(pr_repo, commit_repo, target_dttm, end_dttm,
                        personal_token):

    def load_github_repos():
        commit_api = CommitApi(CommitFactory(), personal_token)
        commit_repo.add_multiple(
            commit_api.get_commit_list(target_dttm, end_dttm))

        pr_api = PullRequestApi(CDDAPullRequestFactory(), personal_token)
        pr_repo.add_multiple(
            pr_api.get_pr_list(target_dttm, end_dttm, merged_only=True))

    github_thread = threading.Thread(
        target=exit_on_exception(load_github_repos))
    github_thread.name = 'WORKER_GIT'
    github_thread.daemon = True
    github_thread.start()

    return github_thread


def get_jenkins_api_data(build_repo):

    def load_jenkins_repo():
        jenkins_api = JenkinsApi(JenkinsBuildFactory())
        build_repo.add_multiple(jenkins_api.get_build_list())

    jenkins_thread = threading.Thread(
        target=exit_on_exception(load_jenkins_repo))
    jenkins_thread.name = 'WORKER_JEN'
    jenkins_thread.daemon = True
    jenkins_thread.start()

    return jenkins_thread


def main_output(by_date, by_build, target_dttm, end_dttm, personal_token,
                include_summary_none, flatten):
    threads = []

    if by_build is not None:
        build_repo = JenkinsBuildRepository()
        threads.append(get_jenkins_api_data(build_repo))

    pr_repo = CDDAPullRequestRepository()
    commit_repo = CommitRepository()
    threads.append(get_github_api_data(pr_repo, commit_repo, target_dttm,
                                       end_dttm, personal_token))

    for thread in threads:
        thread.join()

    if by_date is not None:
        with smart_open(by_date, 'w', encoding='utf8') as output_file:
            build_output_by_date(pr_repo, commit_repo, target_dttm, end_dttm,
                                 output_file, include_summary_none, flatten)

    if by_build is not None:
        with smart_open(by_build, 'w', encoding='utf8') as output_file:
            build_output_by_build(build_repo, pr_repo, commit_repo,
                                  output_file, include_summary_none)


def build_output_by_date(pr_repo, commit_repo, target_dttm, end_dttm,
                         output_file, include_summary_none, flatten):
    # group commits with no PR by date
    commits_with_no_pr = collections.defaultdict(list)
    for commit in commit_repo.traverse_commits_by_first_parent():
        if not pr_repo.get_pr_by_merge_hash(commit.hash):
            commits_with_no_pr[commit.commit_date].append(commit)

    # group PRs by date
    pr_with_summary = collections.defaultdict(list)
    pr_with_invalid_summary = collections.defaultdict(list)
    pr_with_summary_none = collections.defaultdict(list)
    for pr in pr_repo.get_merged_pr_list_by_date(end_dttm, target_dttm):
        if pr.has_valid_summary and pr.summ_type == SummaryType.NONE:
            pr_with_summary_none[pr.merge_date].append(pr)
        elif pr.has_valid_summary:
            pr_with_summary[pr.merge_date].append(pr)
        elif not pr.has_valid_summary:
            pr_with_invalid_summary[pr.merge_date].append(pr)

    # build main changelog output
    offset_dates = (
        end_dttm.date() - timedelta(days=x) for x in itertools.count())
    for curr_date in offset_dates:
        if curr_date < target_dttm.date():
            break
        if curr_date not in pr_with_summary:
            if curr_date not in commits_with_no_pr:
                continue

        print(f"{curr_date}", file=output_file, end='\n\n')

        def sort_by_type(pr):
            return pr.summ_type

        sorted_pr_list_by_category = sorted(pr_with_summary[curr_date],
                                            key=sort_by_type)
        prs_by_type = itertools.groupby(
            sorted_pr_list_by_category, key=sort_by_type)
        for pr_type, pr_list_by_category in prs_by_type:
            if not flatten:
                print(f"    {pr_type}", file=output_file)
            for pr in pr_list_by_category:
                if flatten:
                    print(f"{pr_type} {pr.summ_desc}", file=output_file)
                else:
                    print(f"        * {pr.summ_desc} (by {pr.author} in "
                          f"PR {pr.id})", file=output_file)
            print(file=output_file)

        if curr_date in commits_with_no_pr:
            if not flatten:
                print("    MISC. COMMITS", file=output_file)
            for commit in commits_with_no_pr[curr_date]:
                if not flatten:
                    print(f"        * {commit.message} (by {commit.author} in "
                          f"Commit {commit.hash[:7]})",
                          file=output_file)
                else:
                    print(f"COMMIT {commit.message})", file=output_file)
            print(file=output_file)

        is_included_summary_none = (
            include_summary_none and curr_date in pr_with_summary_none)
        if curr_date in pr_with_invalid_summary or is_included_summary_none:
            if not flatten:
                print("    MISC. PULL REQUESTS", file=output_file)
            for pr in pr_with_invalid_summary[curr_date]:
                if not flatten:
                    print(f"        * {pr.title} (by {pr.author} in "
                          f"PR {pr.id})", file=output_file)
                else:
                    print(f"INVALID_SUMMARY {pr.title})", file=output_file)
            if include_summary_none:
                for pr in pr_with_summary_none[curr_date]:
                    if not flatten:
                        print(f"        * [MINOR] {pr.title} (by {pr.author} "
                              f"in PR {pr.id})", file=output_file)
                    else:
                        print(f"MINOR {pr.title})", file=output_file)
            print(file=output_file)

        print(file=output_file)


def build_output_by_build(build_repo, pr_repo, commit_repo, output_file,
                          include_summary_none):
    # "ABORTED" builds have no "hash" and fucks up the logic here... but just
    # to be sure, ignore builds without hash and changes will be atributed to
    # next build availiable that does have a hash
    def has_hash(build):
        return build.last_hash is not None
    sorted_builds = build_repo.get_all_builds(
        filter_by=has_hash, sort_by=lambda x: -x.build_dttm.timestamp())
    for build in sorted_builds:
        # we need the previous build a hash/date hash
        # to find commits / pull requests that got into the build
        prev_build = build_repo.get_previous_build(build.number, has_hash)
        if prev_build is None:
            break

        try:
            commits = list(commit_repo.get_commit_range_by_hash(
                build.last_hash, prev_build.last_hash))
        except MissingCommitException:
            # we obtained half of the build's commit with our GitHub API
            # request just avoid showing this build's partial data
            break

        print(f'BUILD {build.number} / {build.build_dttm} UTC+0 / '
              f'{build.last_hash[:7]}', file=output_file, end='\n\n')
        if build.last_hash == prev_build.last_hash:
            print(f'  * No changes. Same code as BUILD {prev_build.number}.',
                  file=output_file)
            # I could skip to next build here, but letting the go continue
            # could help spot bugs in the logic the code should not generate
            # any output lines for these Builds.
            #continue

        # I can get PRs by matching Build Commits to PRs by Merge Hash
        # This is precise, but may fail to associate some Commits to PRs
        # leaving few "Summaries" out
        # pull_requests = (pr_repo.get_pr_by_merge_hash(c.hash)
        #                  for c in commits
        #                  if pr_repo.get_pr_by_merge_hash(c.hash))
        # Another option is to find a proper Date range for the Build and find
        # PRs by date; but I found some issues here:
        # * Jenkins Build Date don't end up matching the correct PRs because
        #   git fetch could be delayed like 15mins
        # * Using Build Commit Dates is better, but a few times it incorrectly
        #   match the PR one build later
        #   The worst ofender seem to be merges with message like "Merge
        #   remote-tracking branch 'origin/pr/25005'"
        # build_commit = commit_repo.get_commit(build.last_hash)
        # prev_build_commit = commit_repo.get_commit(prev_build.last_hash)
        # pull_requests = pr_repo.get_merged_pr_list_by_date(
        #     build_commit.commit_dttm + timedelta(seconds=2),
        #     prev_build_commit.commit_dttm + timedelta(seconds=2))

        # I'll go with the safe method, this will show some COMMIT messages
        # instead of the proper Summary from the PR.
        pull_requests = (
            pr_repo.get_pr_by_merge_hash(c.hash) for c in commits
            if pr_repo.get_pr_by_merge_hash(c.hash))

        commits_with_no_pr = [
            c for c in commits if not pr_repo.get_pr_by_merge_hash(c.hash)]

        pr_with_summary = list()
        pr_with_invalid_summary = list()
        pr_with_summary_none = list()
        for pr in pull_requests:
            if pr.has_valid_summary and pr.summ_type == SummaryType.NONE:
                pr_with_summary_none.append(pr)
            elif pr.has_valid_summary:
                pr_with_summary.append(pr)
            elif not pr.has_valid_summary:
                pr_with_invalid_summary.append(pr)

        def sort_by_type(pr):
            return pr.summ_type
        sorted_pr_list_by_category = sorted(pr_with_summary, key=sort_by_type)
        prs_by_type = itertools.groupby(
            sorted_pr_list_by_category, key=sort_by_type)
        for pr_type, pr_list_by_category in prs_by_type:
            print(f"    {pr_type}", file=output_file)
            for pr in pr_list_by_category:
                print(f"        * {pr.summ_desc} (by {pr.author} in "
                      f"PR {pr.id})", file=output_file)
            print(file=output_file)

        if len(commits_with_no_pr) > 0:
            print("    MISC. COMMITS", file=output_file)
            for commit in commits_with_no_pr:
                print(f"        * {commit.message} (by {commit.author} in "
                      f"Commit {commit.hash[:7]})", file=output_file)
            print(file=output_file)

        is_included_summary_none = (
            include_summary_none and len(pr_with_summary_none) > 0)
        if len(pr_with_invalid_summary) > 0 or is_included_summary_none:
            print("    MISC. PULL REQUESTS", file=output_file)
            for pr in pr_with_invalid_summary:
                print(f"        * {pr.title} (by {pr.author} in PR {pr.id})",
                      file=output_file)
            if include_summary_none:
                for pr in pr_with_summary_none:
                    print(f"        * [MINOR] {pr.title} (by {pr.author} in "
                          f"PR {pr.id})", file=output_file)
            print(file=output_file)

        print(file=output_file)


if __name__ == '__main__':
    main_entry(sys.argv)



from ..write_text import write_text


def parse_palette(json, origin):
    if "signs" in json:
        for x in sorted(json["signs"]):
            write_text(json["signs"][x]["signage"], origin,
                       comment="Signage in mapgen palette")



from ..write_text import write_text


def parse_dream(json, origin):
    for msg in json.get("messages", []):
        write_text(msg, origin,
                   comment="Dream of mutation category \"{}\""
                   .format(json["category"]))



from ..helper import get_singular_name
from ..write_text import write_text


def parse_vehicle_part(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Vehicle part name")
    elif "id" in json:
        name = json["id"]

    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of vehicle part \"{}\"".format(name))

    for v in json.get("variants", []):
        id = ""
        label = "Default"
        if "id" in v:
            id = v["id"]
        if "label" in v:
            label = v["label"]
        write_text(label, origin,
                   context="vpart_variants",
                   comment="Label of vehicle part variant \"{}\"".format(id))
        for base in json.get("variants_bases", []):
            based_id = base["id"]
            based_label = base["label"]
            if id != "":
                based_id = base["id"] + "_" + id
            if label != "":
                based_label = base["label"] + " " + label
            # awkward linter dictates 79 chars max, have to line break here
            cmt = "Label of vehicle part variant \"{}\"".format(based_id)
            write_text(based_label, origin,
                       context="vpart_variants",
                       comment=cmt)



#!/usr/bin/env python3
from . import Tokenizer, KnownWords
import unicodedata


def not_in_known_words(word):
    return unicodedata.normalize('NFC', word) not in KnownWords


def spell_check(message):
    words = Tokenizer.findall(message)
    unknowns = filter(not_in_known_words, words)
    return list(unknowns)



import json

from .parser import parsers


def parse_json_object(json, origin):
    """
    Extract strings from the JSON object.
    Silently ignores JSON objects without "type" key.
    Raises exception if the JSON object contains unrecognized "type".
    """
    if "type" in json and type(json["type"]) is str:
        json_type = json["type"].lower()
        if json_type in parsers:
            try:
                parsers[json_type](json, origin)
            except Exception as E:
                print("Exception when parsing JSON data type \"{}\""
                      .format(json_type))
                raise E
        else:
            raise Exception("Unrecognized JSON data type \"{}\""
                            .format(json_type))


def parse_json_file(file_path):
    """Extract strings from the specified JSON file."""
    with open(file_path, encoding="utf-8") as fp:
        json_data = json.load(fp)

    try:
        json_objects = json_data if type(json_data) is list else [json_data]
        for json_object in json_objects:
            parse_json_object(json_object, file_path)
    except Exception:
        print("Error in JSON object\n'{0}'\nfrom file: '{1}'".format(
            json.dumps(json_object, indent=2), file_path))
        raise



from ..helper import get_singular_name
from ..write_text import write_text


def parse_body_part(json, origin):
    # See comments in `body_part_struct::load` of bodypart.cpp about why xxx
    # and xxx_multiple are not inside a single translation object.

    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Name of body part")
    elif "id" in json:
        name = json["id"]

    if "name_multiple" in json:
        write_text(json["name_multiple"], origin, comment="Name of body part")

    if "accusative" in json:
        write_text(json["accusative"], origin,
                   comment="Accusative name of body part")

    if "accusative_multiple" in json:
        write_text(json["accusative_multiple"], origin,
                   comment="Accusative name of body part")

    if "encumbrance_text" in json:
        write_text(json["encumbrance_text"], origin,
                   comment="Encumbrance text of body part \"{}\"".format(name))

    if "heading" in json:
        write_text(json["heading"], origin,
                   comment="Heading of body part \"{}\"".format(name))

    if "heading_multiple" in json:
        write_text(json["heading_multiple"], origin,
                   comment="Heading of body part \"{}\"".format(name))

    if "smash_message" in json:
        write_text(json["smash_message"], origin,
                   comment="Smash message of body part \"{}\"".format(name))

    if "hp_bar_ui_text" in json:
        write_text(json["hp_bar_ui_text"], origin,
                   comment="HP bar UI text of body part \"{}\"".format(name))



from ..write_text import write_text


def parse_skill_display_type(json, origin):
    write_text(json["display_string"], origin,
               comment="Display string for skill display type \"{}\""
               .format(json["id"]))



from ..write_text import write_text


def transform_result(result):
    """
    Canonicalize terrain furniture transform results.
    Transform result may be a single string, a list of strings,
    or a list of list of string and associated probability weights.
    """
    if type(result) is str:
        return result
    elif type(result) is list:
        # Using a list to reserve order and ensure stability across runs
        results = []
        for r in result:
            if type(r) is str and r not in results:
                results.append(r)
            elif type(r) is list and r[0] not in results:
                results.append(r[0])
        return ", ".join(results)


def parse_ter_furn_transform(json, origin):
    if "fail_message" in json:
        write_text(json["fail_message"], origin,
                   comment="Failure message of terrain furniture transform")
    if "terrain" in json:
        for terrain in json["terrain"]:
            if "message" in terrain:
                write_text(terrain["message"], origin,
                           comment="Message after transforming to \"{}\"".
                           format(transform_result(terrain["result"])))
    if "furniture" in json:
        for furniture in json["furniture"]:
            if "message" in furniture:
                write_text(furniture["message"], origin,
                           comment="Message after transforming to \"{}\""
                           .format(transform_result(furniture["result"])))



from .enchant import parse_enchant
from ..helper import get_singular_name
from ..write_text import write_text


def parse_mutation(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Mutation name")

    if "description" in json:
        write_text(json["description"], origin, c_format=False,
                   comment="Description of mutation \"{}\"".format(name))

    if "attacks" in json:
        attacks = json["attacks"]
        if type(attacks) is dict:
            attacks = [attacks]
        for attack in attacks:
            if "attack_text_u" in attack:
                write_text(attack["attack_text_u"], origin,
                           comment="Message when player with mutation "
                           "\"{}\" attacks".format(name))
            if "attack_text_npc" in attack:
                write_text(attack["attack_text_npc"], origin,
                           comment="Message when NPC with mutation "
                           "\"{}\" attacks".format(name))

    if "ranged_mutation" in json:
        if "message" in json["ranged_mutation"]:
            write_text(json["ranged_mutation"]["message"], origin,
                       comment="Message when firing ranged attack with "
                       "mutation \"{}\"".format(name))

    if "spawn_item" in json:
        if "message" in json["spawn_item"]:
            write_text(json["spawn_item"]["message"], origin,
                       comment="Message when spawning item \"{}\" "
                       "with mutation \"{}\""
                       .format(json["spawn_item"]["type"], name))

    if "triggers" in json:
        for arr in json["triggers"]:
            for trigger in arr:
                if "msg_on" in trigger and "text" in trigger["msg_on"]:
                    write_text(trigger["msg_on"]["text"], origin,
                               comment="Trigger message of mutation \"{}\""
                               .format(name))
                if "msg_off" in trigger and "text" in trigger["msg_off"]:
                    write_text(trigger["msg_off"]["text"], origin,
                               comment="Trigger message of mutation \"{}\""
                               .format(name))

    if "variants" in json:
        for variant in json["variants"]:
            variant_name = get_singular_name(variant["name"])
            write_text(variant["name"], origin,
                       comment="Variant name of mutation \"{}\"".format(name))
            write_text(variant["description"], origin,
                       comment="Description of variant " +
                       "\"{1}\" of mutation\"{0}\"".format(name, variant_name))

    if "transform" in json:
        write_text(json["transform"]["msg_transform"], origin,
                   comment="Message when transforming from mutation "
                   " \"{}\" to \"{}\""
                   .format(name, json["transform"]["target"]))

    if "enchantments" in json:
        for enchantment in json["enchantments"]:
            parse_enchant(enchantment, origin)
